#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2019/1/10 2:04 PM
# __author__ = 'Zhou Ran'
u"""
Fetch protein information from uniprot website
"""
import json
from types import SimpleNamespace
from typing import Optional
from xml.parsers.expat import ExpatError

import numpy as np
import requests as rq
import xmltodict
from loguru import logger

from sashimi.conf.DomainSetting import __VALID_DOMAIN_CATEGORY__


class Uniprot(object):
    u"""
    Get domain information from uniprot based uniprot id
    """

    def __init__(self, uniprot_id: str, cds_len: int, fmt="xml", database="uniprot",
                 timeout: int = 10, proxy: Optional[str] = None):
        u"""
        Fetch transcript's translation information based gene id
        :param uniprot_id: the ensemble transcript id works well currently
        :param cds_len: the length of CDS for checking results from uniprot
        :param fmt: default: xml
        :param database: default: uniprot
        """
        self.ui = uniprot_id
        self.cds_len = cds_len
        self.fmt = fmt
        self.database = database
        self.timeout = timeout
        self.proxy = {
            "http": proxy,
            "https": proxy
        }

        self.__valid_fmt = {'txt', 'xml', 'rdf', 'gff', 'fasta'}
        self.__url = "https://rest.uniprot.org"
        self.request_res = self.__request_url__()
        self.guessed_id = self.__guess_protein_id__()
        self.domain = self.__domain_info__()

    def __request_url__(self):
        """
        Request the uniprot url and return
        :return: a Response object from requests
        """
        assert self.fmt in self.__valid_fmt, f"Nonlegal format was found, {self.fmt}"

        request_url = f"{self.__url}/uniprotkb/search?&query={self.ui}&format={self.fmt}"

        try:
            url_response = rq.get(request_url, timeout=self.timeout, proxies=self.proxy)
        except ConnectionError:
            raise f"Failed connect to for {request_url}."
        except rq.exceptions.Timeout:
            raise f"Timeout for {request_url}."

        return url_response

    def __guess_protein_id__(self):
        u"""
        Use xml to parser the response contents and return a dict
        :return: a dict which contained the response results from the given url.
        """

        try:
            # No Domain information
            if self.request_res.text == "":
                return None

            xml_dic = xmltodict.parse(
                self.request_res.text, attr_prefix="", cdata_key=""
            )["uniprot"]["entry"]
            current_uniprot_id = set()

            u"""
            There were more than one protein for a transcript, like ENST00000421241.
            We selected the first protein for downstream visualization because usually was reviewed by uniprot, 
            """

            all_alternative_uniprot_id = []
            if isinstance(xml_dic, list):
                features_length = []
                for sub_domain in xml_dic:
                    features_length.append(
                        int(sub_domain['sequence']['length'])
                    )
                    try:
                        for i in sub_domain["comment"]:
                            if "isoform" in i.keys():
                                all_alternative_uniprot_id.extend(list(map(lambda x: x['id'], i["isoform"])))
                    except (KeyError, AttributeError):
                        continue

                if len(all_alternative_uniprot_id) != 0:
                    features_length = np.array(features_length)
                    length_match_index = np.where(
                        features_length == int(self.cds_len / 3)
                    )[0]
                else:
                    length_match_index = []

                u"""
                Some protein's length was not equal to uniprot, the transcript maybe a non-canonical isoform.  
                """
                if len(length_match_index) != 0:
                    # if multiple, then return the value at minimal index
                    all_alternative_uniprot_id = []
                    xml_dic = xml_dic[min(length_match_index)]
                else:
                    xml_dic = xml_dic[0]

            if int(xml_dic["sequence"]["length"]) != int(self.cds_len / 3):

                if len(all_alternative_uniprot_id) == 0:
                    try:
                        for i in xml_dic["comment"]:
                            if "isoform" in i.keys():
                                all_alternative_uniprot_id = list(map(lambda x: x['id'], i["isoform"]))
                    except KeyError:
                        pass

                if len(all_alternative_uniprot_id) == 0:
                    return None

                for alternative_id in all_alternative_uniprot_id:
                    request_url = f"https://www.ebi.ac.uk/proteins/api/features/{alternative_id}"
                    try:
                        current_uniprot_inf = rq.get(
                            request_url,
                            timeout=self.timeout,
                            proxies=self.proxy
                        )
                        current_uniprot_inf = json.loads(current_uniprot_inf.text)

                        if len(current_uniprot_inf["sequence"]) == self.cds_len / 3:
                            current_uniprot_id.add(current_uniprot_inf['accession'])

                    except ConnectionError:
                        raise f"Failed connect to for {request_url}."
                    except rq.exceptions.Timeout:
                        raise f"Timeout for {request_url}."
            else:
                if isinstance(xml_dic['accession'], list):
                    return xml_dic['accession'][0]
                return xml_dic['accession']

        except ExpatError:
            logger.warning(f"Timeout or no domain information found, id: {self.ui}.")
            return None
        except KeyError:
            # logger.warning(f"Timeout or no domain information found, id: {self.ui}.")
            return None

        # this set may be empty
        if not current_uniprot_id:
            return None

        # if there were multiple protein id, return first.
        return list(current_uniprot_id)[0]

    def __domain_info__(self):
        u"""
        Check the attribution of "feature" in the response results.
        :return: a list which contained feature's attribution
        """
        feature_info = rq.get(f"https://www.ebi.ac.uk/proteins/api/features/{self.guessed_id}",
                              timeout=self.timeout, proxies=self.proxy)

        try:
            feature_info = json.loads(feature_info.text)['features']
            res = []
            if not isinstance(feature_info, list):
                feature_info = [feature_info]

            for sub_feature in feature_info:
                if "description" not in sub_feature.keys():
                    sub_feature["description"] = ""
                else:
                    # Helical; Name=4
                    sub_feature["description"] = sub_feature["description"].split("; ")[0]

                sub_feature["unique_id"] = ",".join(
                    [sub_feature["category"],
                     sub_feature["type"],
                     sub_feature["description"]]
                )

                sub_feature = SimpleNamespace(**sub_feature)
                if sub_feature.category in __VALID_DOMAIN_CATEGORY__:
                    res.append(sub_feature)

            if len(res) == 0:
                return None
            res = sorted(res, key=lambda x: x.unique_id)
            return res
        except KeyError:
            return None


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
u"""

"""


class AxLabel(object):
    def __init__(self, ax, label):
        self.Ax = ax
        self.Label = label

    def __hash__(self):
        return hash(self.Ax)

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
u"""
Created by ygidtu@gmail.com at 2019.12.06
"""


class Junction(object):
    u"""
    Created by ygidtu at 2018.12.19

    This is used to collect information of single junction
    And provide relative position comparison
    """

    __slots__ = ["chromosome", "start", "end", "strand"]

    def __init__(self, chromosome, start, end, strand: str = "+"):
        u"""
        init this class
        :param chromosome: the chromosome name of the given junction
        :param start: the start site of the given junction
        :param end: the end site of the given junction
        :param strand: the strand of the given junction.
        """
        self.chromosome = chromosome
        self.start = int(start)
        self.end = int(end)
        self.strand = strand

        if self.end <= self.start:
            raise ValueError(f"End site({start}) should bigger than start site({end})")

    @property
    def length(self):
        u"""
        :return: int, the length of this junction
        """
        return self.end - self.start

    @classmethod
    def create_junction(cls, string):
        u"""
        create Junction from chr1:1-100:+
        :param string: str, chr1:1-100:+ format or chr1:1-100 also work
        :return:
        """
        string = string.split(":")

        chromosome = string[0]
        start, end = string[1].split("-")

        strand = "+"
        if len(string) > 2:
            strand = string[-1]

        return cls(chromosome=chromosome, start=start, end=end, strand=strand)

    def __hash__(self):
        u"""
        generate hash
        :return:
        """
        return hash((self.chromosome, self.start, self.end, self.strand))

    def __str__(self):
        u"""
        convert junctions to string
        :return:
        """
        return "{chrom}:{start}-{end}".format(
            **{
                "chrom": self.chromosome,
                "start": self.start,
                "end": self.end
            }
        )

    def __gt__(self, other):
        u"""
        greater than
        compare two junction by length
        :param other:
        :return:
        """
        return self.length > other.length

    def __lt__(self, other):
        u"""
        less than
        compare two junction by length
        :param other:a
        :return:
        """
        return self.length < other.length

    def __eq__(self, other):
        u"""
        same length
        :param other:
        :return:
        """
        return self.length == other.length

    def is_overlap(self, other):
        u"""
        whether any overlap with another Junction or GenomicLoci
        :param other:
        :return:
        """

        if self.chromosome != other.chromosome:
            return False

        return self.start < other.end and self.end > other.start

    def is_upstream(self, other):
        u"""
        whether this junction is upstream of other
        :param other:
        :return:
        """
        assert isinstance(other, Junction), "Input should be Junction class"

        if self.chromosome != other.chromosome:
            return self.chromosome < other.chromosome

        return self.end < self.start

    def is_downstream(self, other):
        u"""
        whether this junction is downstream of other
        :param other:
        :return:
        """
        assert isinstance(other, Junction), "Input should be Junction class"

        if self.chromosome != other.chromosome:
            return self.chromosome > other.chromosome

        return self.start > other.end


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
u"""

"""
import os

import pysam
from loguru import logger

from sashimi.base.GenomicLoci import GenomicLoci
from sashimi.file.File import File


class Fasta(File):

    def __init__(self, path: str):
        super().__init__(path)
        assert os.path.exists(path), f"{path} not exists"

    @classmethod
    def create(cls, path: str):
        if not os.path.exists(path + ".fai"):
            logger.warning(f"{path}.fai not exists, try to create it")

            pysam.faidx(path)

        return cls(path)

    def load(self, region: GenomicLoci, **kwargs):
        u"""
        load sequence from a fasta file
        :param region:
        :return:
        """

        if not os.path.exists(self.path + ".fai"):
            logger.warning(f"{self.path}.fai not exists, try to create it")
            pysam.faidx(self.path)

        self.data = {}
        with pysam.FastaFile(self.path) as r:
            for i, j in zip(
                    range(self.region.start, self.region.end + 1),
                    r.fetch(region.chromosome, region.start, region.end + 1)
            ):
                self.data[i] = j
        return self


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
u"""
The parent object of input files
"""
from copy import deepcopy
from typing import Optional, Set, List, Dict


class File(object):
    def __init__(self, path: str):
        self.path = path
        self.data = None
        self.label = ""
        self.region = None
        self.log_trans = "0"
        self.title = ""
        self.is_single_cell = False

    @property
    def chrom(self) -> str:
        if self.region:
            return self.region.chrom
        return ""

    @property
    def start(self) -> int:
        if self.region:
            return self.region.start
        return 0

    @property
    def end(self) -> int:
        if self.region:
            return self.region.end
        return 0

    def load(self, *args, **kwargs):
        return None

    def len(self, scale=1) -> int:
        return len(self.data) / scale if self.data else 0

    def __hash__(self) -> int:
        return hash((self.path, self.label, self.title))

    def __eq__(self, other):
        return self.path == other.path and self.label == other.label and self.title == other.title

    def transform(self):
        if self.data is not None:
            self.data.transform(self.log_trans)


def __set_barcodes__(barcodes: Optional[List[str]]) -> Dict:
    u"""
    separate barcodes by its first character to reduce set size
    :params barcodes: list or set of barcodes
    """
    res = {}

    if barcodes is not None:
        for b in barcodes:
            if b:
                f = b[:min(3, len(b))]

                if f not in res.keys():
                    res[f] = set()

                res[f].add(b)

    return res


class SingleCell(File):
    def __init__(self, path: str, barcodes: Optional[Set[str]] = None, barcode_tag: str = "CB", umi_tag: str = "UB"):

        super().__init__(path)
        self.barcodes = __set_barcodes__(barcodes)
        self.barcode_tag = barcode_tag
        self.umi_tag = umi_tag
        self.is_single_cell = not self.empty_barcode()

    def has_barcode(self, barcode: str) -> bool:
        u"""
        check whether contains barcodes
        :param barcode: barcode string
        """
        if barcode:
            f = barcode[:min(3, len(barcode))]

            temp = self.barcodes.get(f, set())

            return barcode in temp
        return False

    def empty_barcode(self) -> bool:
        u"""
        check whether this bam do not contain any barcodes

        """
        count = 0

        for i in self.barcodes.values():
            count += len(i)

            if count > 0:
                return False

        return True

    def __add__(self, other):
        self.path += other.path

        for i, j in other.barcodes.items():
            if i not in self.barcodes.keys():
                self.barcodes[i] = j
            else:
                self.barcodes[i] |= j

        return self

    def __eq__(self, other) -> bool:
        return self.__hash__() == other.__hash__()

    def copy(self):
        return deepcopy(self)


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*-coding:utf-8 -*-

u"""
Handle the customized junction
"""
import os
import re
from typing import Dict

from sashimi.base.Junction import Junction


def load_custom_junction(path: str) -> Dict[str, Dict[Junction, int]]:
    data = {}
    if os.path.exists(path) and os.path.isfile(path):
        header = []
        with open(path) as r:
            for line in r:
                if line.startswith("#"):
                    continue

                line = line.strip().split()

                if not header:
                    header = line
                else:
                    junc = Junction.create_junction(line[0])
                    for i in range(1, len(header)):
                        sample = header[i]
                        if sample not in data.keys():
                            data[sample] = {}
                        data[sample][junc] = int(line[i])

    return data


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

IMAGE_TYPE = ["reference", "interval", "density", "heatmap", "line", "igv", "hic", "motif"]

COLORS = ["#A0C807", "#0084D1", '#db5f57', '#dbc257', '#91db57', '#57db80', '#57d3db', '#5770db', '#a157db', '#db57b2']
COLORMAP = [
    'Blues', 'Greens', 'Oranges', 'Reds', 'Greys', 'Purples',
    'YlOrBr', 'YlOrRd', 'OrRd', 'PuRd', 'RdPu', 'BuPu',
    'GnBu', 'PuBu', 'YlGnBu', 'PuBuGn', 'BuGn', 'YlGn',
    'binary', 'gist_yarg', 'gist_gray', 'gray', 'bone',
    'viridis', 'plasma', 'inferno', 'magma', 'cividis',
    'pink', 'spring', 'summer', 'autumn', 'winter',
    'cool', 'Wistia', 'hot', 'afmhot', 'gist_heat',
    'copper', 'PiYG', 'PRGn', 'BrBG', 'PuOr', 'RdGy', 'RdBu', 'RdYlBu',
    'RdYlGn', 'Spectral', 'coolwarm', 'bwr', 'seismic'
]

DISTANCE_METRIC = [
    "braycurtis", "canberra", "chebyshev",
    "cityblock", "correlation", "cosine",
    "dice", "euclidean", "hamming",
    "jaccard", "jensenshannon", "kulsinski",
    "kulczynski1", "mahalanobis", "matching",
    "minkowski", "rogerstanimoto", "russellrao",
    "seuclidean", "sokalmichener", "sokalsneath",
    "sqeuclidean", "yule"
]

CLUSTERING_METHOD = ["single", "complete", "average", "weighted", "centroid", "median", "ward"]


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
u"""

"""
from sashimi.base.GenomicLoci import GenomicLoci


class Stroke(object):
    def __init__(self, start: int, end: int, color: str = "red", label: str = ""):
        self.start = start
        self.end = end
        self.color = color
        self.label = label

    @property
    def center(self) -> float:
        return (self.end + self.start) / 2

    @classmethod
    def create(cls, stroke: str, region: GenomicLoci, default_color: str = "red"):
        res = []
        for i in stroke.split(":"):
            i = i.split("@")
            sites = sorted([int(x) - region.start for x in i[0].split("-")])
            if sites[0] < 0:
                sites[0] = 0

            if sites[-1] > len(region):
                sites[-1] = len(region)

            color = default_color
            label = ""
            if len(i) > 1:
                i = i[-1].split("-")
                color = i[0]

                if len(i) > 1:
                    label = i[-1]

            res.append(Stroke(sites[0], sites[-1], color, label))
        return res


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2022/5/8 10:04 AM
u"""
Fetch CDS information from a gtf then mapping to genomic position
"""
from collections import defaultdict, namedtuple
from typing import Optional

from loguru import logger
from pysam import asGTF
from pysam.libctabix import Tabixfile

from sashimi.base.CoordinateMap import CoordinateMapper
from sashimi.base.GenomicLoci import GenomicLoci
from sashimi.base.Transcript import Transcript
from sashimi.base.pyUniprot import Uniprot


class CdsProtein(GenomicLoci):
    u"""
    Re-collect cds information
    """

    def __init__(
            self,
            chromosome: str,
            start: int,
            end: int,
            strand: str,
            cds: dict,
            gene_id: set,
            transcript_id: set,
            proxy: Optional[str] = None,
            timeout: int = 10
    ):
        u"""
        Re-iter the gtf and collection all CDS information from gtf, not truncated genomic region at here
        :param chromosome: the chromosome id of the  given genes
        :param start: the minima start site of the given genes
        :param end: the maxima start site of the given genes
        :param strand: strand of gene
        :param cds: a list of Transcript object
        :param gene_id: a set of gene_id
        :param transcript_id: a set of transcript_id
        """
        super().__init__(
            chromosome=chromosome,
            start=start,
            end=end,
            strand=strand)
        self.proxy = proxy
        self.timeout = timeout
        self.cds = cds
        self.pep = self.__init_protein__() if cds is not None else None
        self.gene_id = gene_id
        self.transcript_id = transcript_id

    def __len__(self):
        return len(self.pep)

    def __add__(self, other):
        assert isinstance(other, CdsProtein), "only CdsProtein and CdsProtein could be added"
        self.gene_id.update(other.gene_id)
        self.transcript_id.update(other.transcript_id)
        self.cds.update(other.cds)
        self.pep.update(other.pep)
        self.region = GenomicLoci(
            chromosome=self.chromosome,
            start=min(self.start, other.start),
            end=max(self.end, other.start),
            strand="."
        )

    # @property
    # def truncated_pep(self):
    #
    #     return

    @classmethod
    def __re_iter_gtf__(
            cls,
            gtf_tabix: Tabixfile,
            chromosome: str,
            transcript_id: set,
            gene_id: set,
            strand: str = "*",
            proxy: Optional[str] = None,
            timeout: int = 10
    ):
        u"""
        Generate the CdsProtein object
        :param gtf_tabix: a Tabix file object generated by pysam
        :param chromosome: the chromosome id of the given genes
        :param transcript_id: the set of transcript_id to fetch CDS region
        :param gene_id: the set of gene_id to fetch CDS region, and also used to stop the iteration
        :param strand: the strand of gene
        :return:
        """
        relevant_exon_iterator = gtf_tabix.fetch(reference=chromosome, parser=asGTF())

        boundary_for_iter = []

        cds_dict = defaultdict(list)

        for line in relevant_exon_iterator:

            if strand != "*" and line.strand != strand:
                continue
            if "transcript_id" not in line.keys():
                continue

            try:
                if len(boundary_for_iter) == len(gene_id) and \
                        line.end > max(boundary_for_iter):
                    break

                # update the lower boundary to stop the iter.
                if line.gene_id in gene_id:
                    boundary_for_iter.append(line.end)

                if line.transcript_id in transcript_id and \
                        line.feature == "CDS":
                    cds_dict[line.transcript_id].append(line)

            except KeyError as err:
                logger.warning(err)

        cds_contain = {}
        start_site = float('Inf')
        end_site = float('-Inf')

        if len(cds_dict) != 0:

            for current_trans_id in cds_dict.keys():
                current_exon_list = cds_dict[current_trans_id]
                current_start = min(map(lambda x: x.start + 1, current_exon_list))
                current_end = max(map(lambda x: x.end, current_exon_list))

                cds_contain[current_trans_id] = Transcript(
                    chromosome=chromosome,
                    start=current_start,
                    end=current_end,
                    strand=current_exon_list[0].strand,
                    exons=current_exon_list,
                    gene=current_exon_list[0].gene_name \
                        if "gene_name" in current_exon_list[0].keys() else "",
                    gene_id=current_exon_list[0].gene_id \
                        if "gene_id" in current_exon_list[0].keys() else "",
                    transcript=current_exon_list[0].transcript_name \
                        if "transcript_name" in current_exon_list[0].keys() else "",
                    transcript_id=current_exon_list[0].transcript_id \
                        if "transcript_id" in current_exon_list[0].keys() else ""
                )
                start_site = current_start if current_start < start_site else start_site
                end_site = current_end if current_end > end_site else end_site

            return cls(
                chromosome=chromosome,
                start=start_site,
                end=end_site,
                strand=strand,
                cds=cds_contain,
                gene_id=gene_id,
                transcript_id=transcript_id,
                proxy=proxy,
                timeout=timeout
            )

        return None

    def __init_protein__(self):
        u"""
        Initiate the CdsProtein object
        :return:
        """

        protein_info = defaultdict(list)
        Domain_region = namedtuple(
            "DomainGenomicRegion",
            ["category", "type", "description", "unique_id", "start", "end"]
        )

        for current_transcript_id in self.cds.keys():
            current_pep = Uniprot(
                uniprot_id=current_transcript_id,
                cds_len=len(self.cds[current_transcript_id]),
                proxy=self.proxy, timeout=self.timeout
            )

            if not current_pep.domain:
                protein_info[current_transcript_id] = []
                continue

            coord_mapper = CoordinateMapper(
                coordinates_list=self.cds[current_transcript_id].exon_list,
                strand=self.cds[current_transcript_id].strand
            )

            domain_res = defaultdict(list)
            for domain in current_pep.domain:
                current_unique_id = domain.unique_id
                pep_genomic_coord = coord_mapper.pep_to_cds(
                    int(domain.begin),
                    int(domain.end)
                )

                # keep the protein genomic coordinate into a list
                current_domain_res = []
                for sub_domain_start, sub_domain_end in pep_genomic_coord.se:
                    current_domain_res.append(
                        Domain_region._make(
                            [domain.category,
                             domain.type,
                             domain.description,
                             current_unique_id,
                             sub_domain_start,
                             sub_domain_end
                             ]
                        )
                    )
                domain_res[current_unique_id].extend([current_domain_res])

            for domain_unique_id, domain_list in domain_res.items():
                start_site = min([
                    min(map(lambda x: x.start, i)) for i in domain_list
                ])

                end_site = max([max(map(lambda x: x.end, i)) for i in domain_list])

                protein_info[current_transcript_id].append(
                    Transcript(
                        chromosome=self.chromosome,
                        start=start_site,
                        end=end_site,
                        strand=self.cds[current_transcript_id].strand,
                        exons=domain_list,
                        gene=domain_list[0][0].unique_id,
                        domain_type=domain_list[0][0].type,
                        domain_description=domain_list[0][0].description,
                        domain_category=domain_list[0][0].category,
                        category="protein"
                    )
                )

        return protein_info


if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2019/1/11 2:55 PM

u"""
Convert aa position into genomic coordinate, Ran zhou.

"""
from itertools import islice
from typing import Optional

import numpy as np


class Coordinate(object):
    u"""
    A Coordinate object for genomic regions.
    """

    def __init__(self, coordinate_list: list, strand: str = "*"):
        u"""
        Set genomic coordinates
        :param coordinate_list: a nested tuple of list
        :param strand: a strands of given coordinate object
        """
        self.strand = strand
        self.se = self.__fmt_exons__(coordinate_list)
        assert len(self.location_list) == len(set(self.location_list)), \
            f"Overlapped regions were found in {self.se}"

    @staticmethod
    def __fmt_exons__(coordinate_list: list) -> list:
        u"""
        Format and sort exon list
        :param coordinate_list: a nested tuple of list. like [('5','6'),('1','4')]
        :return: a nested tuple of list. like [(1,4), (5,6)]
        """
        # sorting coordinate based on first of location.
        formatted_coordinate_list = list(map(lambda x: tuple(map(int, x)), coordinate_list))
        formatted_coordinate_list.sort(key=lambda x: x[0])
        return formatted_coordinate_list

    @staticmethod
    def __get_s_or_e__(coordinate_list: list, index: int) -> list:
        u"""
        Get start or end site for each given coordinates
        :param coordinate_list: a nested tuple of list, like [(1,3),(5,6)]
        :param index: the index of tuple, 0 for the left site and 1 for the right end site.
        :return: a list which contained the start or end sites.
        """
        return list(map(lambda x: x[index], coordinate_list))

    @property
    def start(self):
        return self.__get_s_or_e__(coordinate_list=self.se, index=0)

    @property
    def end(self):
        return self.__get_s_or_e__(coordinate_list=self.se, index=1)

    @property
    def introns(self):
        u"""
        Set intronic regions for each coordinate object
        :return: a nested tuple of list which contained intronic coordinates.
        """
        if len(self.se) == 1:
            return None
        else:
            introns_list = []
            for left_exon, right_exon in self.__slide_window__(self.se, num_of_chunk=2):
                introns_list.append(tuple(
                    [left_exon[1] + 1,
                     right_exon[0] - 1]
                ))
            return introns_list

    @staticmethod
    def __slide_window__(nested_list: list, num_of_chunk: int):
        u"""
        A sliding window to slice the given list
        :param nested_list: a nested tuple of list, like [(1,3),(5,6)]
        :param num_of_chunk:  num of element for each chunk.
        :return: 
        """
        nested_list = iter(nested_list)
        chunked_list = list(islice(nested_list, num_of_chunk))
        if len(chunked_list) == num_of_chunk:
            yield chunked_list
        for elem in nested_list:
            result = chunked_list[1:] + list((elem,))
            yield result

    @classmethod
    def __flatten__(cls, nested_list: list):
        u"""
        Flatten the nested list
        :param nested_list: a nested tuple of list, like [(1,3),(5,6)]
        :return:
        """
        for sub_list in nested_list:
            if hasattr(sub_list, "__iter__") and not isinstance(sub_list, str):
                for sub_el in cls.__flatten__(sub_list):
                    yield sub_el
            else:
                yield sub_list

    @property
    def location_list(self) -> np.ndarray:
        u"""
        Get a list which contained all position.
        For example, an exon list `[(10, 15), (20,25)]` as input,
        [10, 11, 12, 13, 14, 15, 20, 21, 22, 23, 24, 25] will return
        :return: a list which contained all position
        """
        # Add 1 offset because of 0-based coordinate
        position_list = list(self.__flatten__(list(map(
            lambda x: range(x[0], x[1] + 1), self.se
        ))))

        if self.strand == '-':
            return np.array(position_list[::-1])

        return np.array(position_list)

    @property
    def pep_index(self) -> np.ndarray:
        u"""
        Relative position of pep coordinates
        :return: 
        """
        return np.array(list(map(lambda x: int(x / 3), range(len(self.location_list)))))

    @property
    def cds_index(self) -> np.ndarray:
        u"""
        Relative position of cds coordinates
        :return:
        """
        return np.array(list(range(len(self.location_list))))

    @staticmethod
    def __group_consecutive_value__(location_list: np.ndarray, strand: str) -> list:
        u"""
        group the consecutive value into a list
        :param location_list: a list of location site
        :param strand: the strand of the current list to group
        :return:
        """
        offset = -1 if strand == "-" else 1
        group_ids = np.concatenate(([0], (np.diff(location_list) != offset).cumsum()))

        grouped_truncated_location_array = \
            np.split(
                location_list,
                np.unique(group_ids, return_counts=True)[1].cumsum().tolist()
            )

        return grouped_truncated_location_array

    @classmethod
    def init_from_location_list(cls, truncated_location_array: np.ndarray, strand: str):
        u"""
        init class based on the list of location
        :param truncated_location_array: truncated location array
        :param strand: the strand of the current list
        :return:
        """
        __coordinate_list = []
        for sub_array in cls.__group_consecutive_value__(truncated_location_array, strand):
            if len(sub_array) == 0:
                continue

            __coordinate_list.append(
                tuple(
                    [
                        min(sub_array),
                        max(sub_array)
                    ]
                )
            )
        return cls(__coordinate_list, strand)


class CoordinateMapper(Coordinate):
    u"""
    Convert positions between CDS and protein coordinates.
    TODO: Add cds to pep coordinate?
    """

    def __init__(self, coordinates_list, strand: str):
        u"""
        Set genomic coordinates to be used for mapping
        :param coordinates_list: a nested tuple of list
        :param strand: a strands of given coordinate object
        """
        super().__init__(coordinate_list=coordinates_list, strand=strand)

    def pep_to_cds(self, pep_start: int, pep_end: Optional[int] = None):
        u"""
        Convert pep position into genomic position
        :param pep_start: the start position of pep
        :param pep_end: the end position of pep, if None, the start site is equal to end site
        :return:
        """
        # change 1-based pep coordinate into 0-based coordinate
        pep_start = pep_start - 1
        pep_end = pep_start if pep_end is None else pep_end - 1

        start_ind = np.where(self.pep_index == pep_start)[0]
        end_ind = np.where(self.pep_index == pep_end)[0]

        cds_left_index, cds_right_index = start_ind[0], end_ind[-1] + 1

        return self.init_from_location_list(
            self.location_list[cds_left_index:cds_right_index],
            self.strand)


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
u"""
Created by ygidtu@gmail.com at 2019.12.06

Changelog:
    1. remove attributes
"""
from typing import List

from sashimi.base.GenomicLoci import GenomicLoci


class Transcript(GenomicLoci):
    u"""
    Created by ygidtu at 2018.12.21

    A class inherit from GenomicLoci, to collect transcript information
    """

    __slots__ = [
        "gene",
        "gene_id",
        "transcript",
        "transcript_id",
        "exons",
        "category",
        "domain_category",
        "domain_type",
        "domain_description",
        "plot_intron"
    ]

    def __init__(
            self,
            chromosome: str,
            start: int,
            end: int,
            strand: str,
            exons: list,
            gene: str = "",
            gene_id: str = "",
            transcript: str = "",
            transcript_id: str = "",
            category: str = "exon",
            domain_category: str = "",
            domain_type: str = "",
            domain_description: str = ""
    ):
        u"""
        :param chromosome:
        :param start:
        :param end:
        :param strand:
        :param exons: A list of pysam.GTFProxy if category was exon, A nested tuple of list if category was protein
        :param gene: gene name when category is exon,such as "SAMD11"; domain's description when category is domain such as "Disordered"
        :param gene_id: gene id, such as "ENSG00000187634"
        :param transcript: transcript name, such as "SAMD11-011"
        :param transcript_id: transcript id, such as "ENST00000420190"
        :param category: exon or protein or interval
        :param domain_category: category of domain
        :param domain_description: description of domain
        :param domain_type: if category is protein, the type information of the given domain
        """

        super().__init__(
            chromosome=chromosome,
            start=start,
            end=end,
            strand=strand
        )
        self.transcript = transcript
        self.transcript_id = transcript_id
        self.gene = gene
        self.gene_id = gene_id
        self.exons = sorted(exons)
        self.category = category
        self.domain_category = domain_category
        self.domain_type = domain_type
        self.domain_description = domain_description

    @property
    def exon_list(self):

        exon_nested_lst = []
        for i in self.exons:
            exon_nested_lst.append(([i.start + 1, i.end]))
        return exon_nested_lst

    def __str__(self):
        exons_str = []
        for i in self.exons:
            if isinstance(i, list):
                """
                2022.07.05
                Domain setting
                """
                exons_str.append("|".join(map(lambda x: f"{x.start}-{x.end}", i)))

            else:
                exons_str.append("{}-{}".format(i.start, i.end))

        return "{}:{}-{}:{} {} {}".format(
            self.chromosome,
            self.start,
            self.end,
            self.strand,
            self.transcript,
            "|".join(exons_str)
        )

    def __len__(self):
        return sum(map(lambda x: x[1] - x[0] + 1, self.exon_list))

    def __hash__(self):
        exons = sorted([str(x.__hash__()) for x in self.exons])
        return hash((self.chromosome, self.start, self.end, self.strand, " ".join(exons)))

    def ids(self) -> List[str]:
        return [self.transcript, self.transcript_id, self.gene, self.gene_id]


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# all the drawing functions in this file assume that
# the coordinates of the svg file have been transformed to cartesian coordinates
class RGB(object):
    def __init__(self, red=255, green=255, blue=255):
        if red > 255 or red < 0 or green > 255 or green < 0 or blue > 255 or blue < 0:
            raise Exception("Invalid color")
        else:
            self.red = int(red)
            self.green = int(green)
            self.blue = int(blue)

    @classmethod
    def from_hex_string(cls, hex_string):
        if len(hex_string) == 7:
            hex_string = hex_string[1:]

        if len(hex_string) == 6:
            red = int(hex_string[0:2], 16)
            green = int(hex_string[2:4], 16)
            blue = int(hex_string[4:], 16)
            return cls(red, green, blue)
        else:
            raise Exception("not a valid hexadecimal color")

    def __str__(self):
        return 'rgb({0}, {1}, {2})'.format(self.red, self.green, self.blue)


def draw_line(svg_file, x1, y1, x2, y2, thickness=10, color=RGB(0, 0, 0)):
    svg_line = '<line x1="{0}" y1="{1}" x2="{2}" y2="{3}" style="stroke: {4}; stroke-width: {5}"/> \n'.format(x1, y1,
                                                                                                              x2, y2,
                                                                                                              color,
                                                                                                              thickness)
    svg_file.write(svg_line)


def draw_line_polar(svg_file, x_origin, y_origin, length, angle, thickness=10, color=RGB(0, 0, 0)):
    svg_line = '<line x1="{0}" y1="{1}" x2="{2}" y2="{3}" style="stroke: {4}; stroke-width: {5}" transform="rotate({6},{0},{1})"/> \n'.format(
        x_origin, y_origin, x_origin + length, y_origin, color, thickness, angle)
    svg_file.write(svg_line)


def draw_bezier(svg_file, x1, y1, x2, y2, controlX, controlY, color=RGB(0, 0, 0), thickness=1):
    svg_bezier = '<path d = "M{0},{1} Q{4},{5} {2},{3}" fill="none" style="stroke: {6}; stroke-width: {7}"/> \n'.format(
        x1, y1, x2, y2, controlX, controlY, color, thickness)
    svg_file.write(svg_bezier)


def draw_text(svg_file, words, x, y, size, angle=0, color=RGB(0, 0, 0)):
    svg_text = '<text x="{0}" y="{1}" style="fill:{2}" font-size="{3}" font-family="sans-serif" transform="scale(1,-1) rotate({5},{0},{1})" text-anchor="middle">{4}</text>\n'.format(
        x, -y, color, size, words, angle)
    svg_file.write(svg_text)


def draw_text_left(svg_file, words, x, y, size, angle=0, color=RGB(0, 0, 0)):
    svg_text = '<text x="{0}" y="{1}" style="fill:{2}" font-size="{3}" font-family="sans-serif" transform="scale(1,-1) rotate({5},{0},{1})" text-anchor="left">{4}</text>\n'.format(
        x, -y, color, size, words, angle)
    svg_file.write(svg_text)


def draw_multiline_text(svg_file, label, x, y, size, color=RGB(0, 0, 0)):
    words_list = label.split('\n')

    svg_file.write(
        '<text y="{0}" style="fill:{1}" font-size="{2}" font-family="sans-serif" transform="scale(1,-1)" text-anchor="middle">'.format(
            -y, color, size))
    for i in range(len(words_list)):
        if i == 0:
            svg_file.write('<tspan x="{0}">{1}</tspan>'.format(x, words_list[i]))
        else:
            svg_file.write('<tspan x="{0}" dy="{1}">{2}</tspan>'.format(x, size, words_list[i]))
    svg_file.write('</text>')


def draw_rectangle(svg_file, x, y, x_dim, y_dim, fill_color):
    svg_rect = '<rect x="{0}" y="{1}" width="{2}" height="{3}" style="fill:{4}; stroke-width:0.01; stroke:{4}"/>\n'.format(
        x, y, x_dim, y_dim, fill_color);
    svg_file.write(svg_rect)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
u"""
Generate object for plotting reads like IGV track
"""
import os.path
import sys
from typing import Optional, List, Dict, Union

import numpy as np
import pandas as pd
import pysam
from loguru import logger

from sashimi.base.CoordinateMap import Coordinate
from sashimi.base.GenomicLoci import GenomicLoci
from sashimi.base.Readder import Reader
from sashimi.file.File import File

np.warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning)


class Reads(GenomicLoci):

    def __init__(self,
                 chromosome: str,
                 start: str,
                 end: str,
                 strand: str,
                 id: str,
                 exons: list,
                 introns: list,
                 polya_length: float = -1.0,
                 m6a: float = -1.0,
                 features: Optional[list] = None):
        u"""
        Fetch information of each read from bam file, for stacked plot like IGV
        :param chromosome: the chromosome id of the given read
        :param start: the start site of the given read
        :param end: the end site of the given read
        :param strand: the strand of the given read
        :param id: the name of the given read
        :param exons: a list of GenomicLoci obj which contains region of exon
        :param introns: a list of GenomicLoci obj which contains region of intron
        :param polya_length: length of polya
        :param features: current support m6a site and polya length, like below
        """

        super().__init__(
            chromosome=chromosome,
            start=start,
            end=end,
            strand=strand
        )

        self.exons = self.__collapse_read__(sorted(exons))
        self.introns = self.__collapse_read__(sorted(introns))
        self.polya_length = polya_length
        self.m6a = m6a
        self.features = features
        self.id = id

    @property
    def exon_list(self):
        u"""
        return a nested list which contains exon regions
        1-based coordinate
        :return: a nested list
        """

        exon_nested_lst = []
        for i in self.exons:
            exon_nested_lst.append(([i.start, i.end]))
        return exon_nested_lst

    @property
    def intron_list(self):
        u"""
        return a nested list which contains intronic regions
        1-based coordinate
        :return: a nested list
        """

        intron_nested_lst = []
        for i in self.introns:
            intron_nested_lst.append(
                ([i.start, i.end])
            )
        return intron_nested_lst

    def __len__(self):
        return sum(map(lambda x: x[1] - x[0] + 1, self.exon_list))

    def __str__(self):
        exons_str = []
        for i in self.exons:
            exons_str.append("{}-{}".format(i.start, i.end))

        return "{}:{}-{}:{} {} {}".format(
            self.chromosome,
            self.start,
            self.end,
            self.strand,
            self.id,
            "|".join(exons_str)
        )

    def to_dict(self) -> Dict:
        u"""
        return a dict for generate DataFrame object
        :return:
        """
        return {
            "chromosome": self.chromosome,
            "start": self.start,
            "end": self.end,
            "strand": self.strand,
            "polya_length": self.polya_length,
            "m6a": self.m6a
        }

    @staticmethod
    def __collapse_read__(genomic_list: List[GenomicLoci]) -> List[GenomicLoci]:
        u"""
        Collapse the consecutive region, because of the coordinate shift of matplotlib
        :param genomic_list: a list of GenomicLoci
        :return:
        """

        if len(genomic_list) <= 1:
            return genomic_list

        list_return = []
        current_gl = None

        for gl in genomic_list:
            if not current_gl:
                current_gl = gl
            else:
                if current_gl.end + 1 == gl.start:
                    current_gl = GenomicLoci(
                        chromosome=current_gl.chromosome,
                        start=current_gl.start,
                        end=gl.end,
                        strand=current_gl.strand,
                        name=current_gl.name
                    )
                else:
                    list_return.append(current_gl)
                    current_gl = gl

        list_return.append(current_gl)
        return list_return


class ReadSegment(File):

    def __init__(
            self,
            path: str,
            label: str = "",
            meta: Optional[pd.DataFrame] = None,
            region: Optional[GenomicLoci] = None,
            library: str = "fru",
            deletion_ignore: Optional[int] = True,
            del_ratio_ignore: float = .5,
            features: Optional[dict] = None,
            exon_focus: Optional[str] = None,
            is_bed: bool = False
    ):
        u"""
        init a class for store the information for IGV-like plot
        :param path: the path of the input file
        :param label: the label of current track
        :param meta: the meta info of all reads which was used to sort the reads
        :param region: the region for plotting
        :param library: the library category of the input file
        :param deletion_ignore: whether to ignore the deletion region, default: Ture
        :param del_ratio_ignore: the length of deletion must below to del_ratio_ignore * length of alignment
        :param features: default: features={"m6a": "ma","real_strand": "rs","polya": "pa"}
        :param exon_focus: exon to focus, like start1-end1,start2-end2
        :param is_bed: bed file for generating igv-like reads.
        """
        super().__init__(path)

        self.features = None
        assert library in ["frf", "frs", "fru"], \
            "Illegal library name."

        self.library = library
        self.data = []
        self.meta = meta
        self.region = region
        self.label = label
        self.deletion_ignore = deletion_ignore
        self.del_ratio_ignore = del_ratio_ignore
        self.features = features
        self.is_bed = is_bed
        self.exon_focus = set(map(lambda x: x.strip(), exon_focus.split(','))) if exon_focus else exon_focus

    @classmethod
    def create(
            cls,
            path: str,
            label: str = "",
            library: str = "fru",
            deletion_ignore: Optional[int] = True,
            del_ratio_ignore: float = .5,
            features: Optional[dict] = None,
            exon_focus: Optional[str] = None

    ):
        u"""
        Load each reads and its strand, features.
        m6a tag support the genomic site
        polya tag support length of polya. if polya tag is available, then read_strand (rl) is also essential

        :param path:
        :param label:
        :param library:
        :param deletion_ignore: ignore the deletion length
        :param del_ratio_ignore: ignore the deletion length which calculated by mapped length * ratio
        :param features: support m6a and polyA length from bam tag.
        like {"m6a": "ma", "polya": "pa", "real_strand": "rs"}
        :param exon_focus: exon to focus, like start1-end1,start2-end2
        :return:
        """
        # if bam file then check the index file of the given file
        if path.endswith("bam") and not os.path.exists(path + ".bai"):
            pysam.index(path)

        is_bed = False

        # if bed file
        if "bed" in path:
            is_bed = True
            if not os.path.exists(path + ".tbi"):
                try:
                    path = pysam.tabix_index(path, preset="bed", force=True)
                except OSError as e:
                    logger.error(f"Failed to build index for {path}. \n {e} \n"
                                 f"Please sort and index your bed file by "
                                 f"`bedtools sort -i {path} | bgzip > {path}.gz`")
                    sys.exit(0)
                    # path_new = re.sub(".bed.gz$", "", path) + 'sorted.bed.gz'
                    #
                    # Reference.sort_gtf(input_gtf=path,
                    #                    output_gtf=path_new)
                    # path = path_new

        return cls(
            path=path,
            label=label,
            library=library,
            deletion_ignore=deletion_ignore,
            del_ratio_ignore=del_ratio_ignore,
            features=features,
            exon_focus=exon_focus,
            is_bed=is_bed
        )

    def get_index(self):
        u"""
        get a nested list which presents the order of plot
        :return:
        """
        assert self.meta is not None, f"Not found meta information, please `load` first."
        for ind in self.meta.groupby(['y_loci'])['list_index'].apply(list).tolist()[::-1]:
            yield ind

    def set_region(self,
                   chromosome,
                   start,
                   end,
                   strand):
        u"""
        set the plotting region
        :param chromosome: the name of chromosome
        :param start: the start of the plotting region
        :param end: the end of the plotting region
        :param strand: stand of the plotting region
        :return:
        """
        self.region = GenomicLoci(
            chromosome=chromosome,
            start=start,
            end=end,
            strand=strand
        )

    @staticmethod
    def df_sort(dfs: pd.DataFrame) -> Optional[pd.DataFrame]:
        u"""
        sorting the dataframe to generate plot index,
        copy from jinbu jia
        :param dfs: a pd.DataFrame object
        :return: pd.DataFrame
        """

        y_loci = []
        have_overlap_regions = []
        now_max_x = 0
        height = 1
        y_min = 1
        dfs_lst = []
        for _, df in dfs.groupby('exon_group'):
            for index, row in df.iterrows():
                start = row["start"]
                end = row["end"]

                if start > now_max_x:
                    if y_min != 0:
                        y_min += 1
                    y_max = height
                    now_max_x = end
                    have_overlap_regions = [[1, y_max, now_max_x]]
                else:
                    for d in have_overlap_regions:
                        if d[2] < start:
                            d[2] = start - 1
                    if len(have_overlap_regions) > 1:
                        new_have_overlap_regions = [have_overlap_regions[0]]
                        for d in have_overlap_regions[1:]:
                            if d[2] == new_have_overlap_regions[-1][2]:
                                new_have_overlap_regions[-1][1] = d[1]
                            else:
                                new_have_overlap_regions.append(d)
                        have_overlap_regions = new_have_overlap_regions
                    have_insert = False
                    for (i, d) in enumerate(have_overlap_regions):
                        x1, x2, x3 = d
                        if x3 < start and (x2 - x1 + 1) >= height:
                            have_insert = True
                            y_min = x1
                            y_max = y_min + height - 1
                            if y_max != x3:
                                d[0] = y_max + 1
                                have_overlap_regions.insert(i, [x1, y_max, end])
                            else:
                                d[2] = end
                            break
                    if not have_insert:
                        y_min = have_overlap_regions[-1][1] + 1 if have_overlap_regions else 1
                        y_max = y_min + height - 1
                        have_overlap_regions.append([y_min, y_max, end])
                    if end > now_max_x:
                        now_max_x = end
                y_loci.append(y_min)
            df["y_loci"] = y_loci
            dfs_lst.append(df)
            y_loci = []

        if len(dfs_lst) <= 0:
            logger.error(f"There is no any read segments")
            return None

        return pd.concat(dfs_lst)

    def load_bed(self):
        try:
            for rec in Reader.read_gtf(self.path, region=self.region, bed=True):
                exon_bound = []
                intron_bound = []
                current_start = int(rec[1])
                current_end = int(rec[2])
                if len(rec) > 3:
                    current_id = rec[3]
                else:
                    current_id = "NoID"

                if len(rec) != 12:
                    exon_bound.append(
                        GenomicLoci(
                            chromosome=self.region.chromosome,
                            start=current_start + 1,
                            end=current_end,
                            strand=self.region.strand,
                            name="exon"
                        )
                    )
                else:

                    block_sizes = [int(x) for x in rec[10].split(",") if x]
                    block_starts = [int(x) for x in rec[11].split(",") if x]

                    read = []
                    for i in range(len(block_starts)):
                        exon_bound.append(
                            GenomicLoci(
                                chromosome=self.region.chromosome,
                                start=current_start + 1 + block_starts[i],
                                end=current_start + 1 + block_starts[i] + block_sizes[i] - 1,
                                strand=self.region.strand,
                                name="exon"
                            )
                        )
                    for pre_exon, next_exon in Coordinate.__slide_window__(exon_bound, 2):
                        intron_bound.append(
                            GenomicLoci(
                                chromosome=self.region.chromosome,
                                start=pre_exon.end + 1,
                                end=next_exon.start - 1,
                                strand=self.region.strand,
                                name="intron"
                            )
                        )

                read = Reads(
                    chromosome=self.region.chromosome,
                    start=min(map(lambda x: x.start, exon_bound)),
                    end=max(map(lambda x: x.end, exon_bound)),
                    strand=self.region.strand,
                    id=current_id,
                    exons=exon_bound,
                    introns=intron_bound,
                    polya_length=-1,
                    m6a=-1,
                    features=[]
                )
                if read.start < self.region.start or read.end > self.region.end:
                    continue

                self.data.append(read)

        except IOError as err:
            logger.error('There is no .bed file at {0}'.format(self.path))
            logger.error(err)
        except ValueError as err:
            logger.error(self.path)
            logger.error(err)

    def load_bam(self):
        try:
            for read, _ in Reader.read_bam(self.path, self.region):
                if read.reference_start < self.region.start or read.reference_end > self.region.end:
                    continue

                if not self.deletion_ignore:
                    current_ignore_num = min(
                        [self.deletion_ignore, read.query_alignment_length * self.del_ratio_ignore])
                else:
                    current_ignore_num = np.Inf

                exon_bound = []
                intron_bound = []

                start = read.reference_start + 1
                for c, l in read.cigar:
                    if c == 0:  # for match
                        exon_bound.append(
                            GenomicLoci(
                                chromosome=self.region.chromosome,
                                start=start,
                                end=start + l - 1,
                                strand=self.region.strand,
                                name="exon"
                            )
                        )
                        start += l

                    elif c == 1:  # for insert
                        continue

                    elif c == 2:  # for del
                        if l <= current_ignore_num:
                            exon_bound.append(
                                GenomicLoci(
                                    chromosome=self.region.chromosome,
                                    start=start,
                                    end=start + l - 1,
                                    strand=self.region.strand,
                                    name="exon"
                                )
                            )
                        start += l

                    elif c == 3:  # for intron
                        intron_bound.append(
                            GenomicLoci(
                                chromosome=self.region.chromosome,
                                start=start,
                                end=start + l - 1,
                                strand=self.region.strand,
                                name="intron"
                            )
                        )
                        start += l

                    elif c == 4:  # soft clip
                        continue

                    else:
                        continue

                features_list = []

                # for m6a
                m6a_loci = -1
                polya_length = -1
                if self.features:
                    if read.has_tag(self.features["m6a"]):
                        # multiple m6a support
                        m6a_loci_list = list(
                            map(float, [loci.strip() for loci in read.get_tag(self.features["m6a"]).split(',')])
                        )
                        for m6a_loci in m6a_loci_list:
                            features_list.append(GenomicLoci(
                                chromosome=self.region.chromosome,
                                start=m6a_loci,
                                end=m6a_loci,
                                strand="*",
                                name="m6a"
                            ))

                    if read.has_tag(self.features["real_strand"]) and read.has_tag(self.features["polya"]):
                        polya_length = float(read.get_tag(self.features["polya"]))
                        real_strand = read.get_tag(self.features["real_strand"])
                        assert read.get_tag(self.features["real_strand"]) in {"+", "-", "*"}, \
                            f"strand should be *, + or -, not {real_strand}"
                        if real_strand == "+":
                            features_list.append(
                                GenomicLoci(
                                    chromosome=self.region.chromosome,
                                    start=read.reference_end + 1,
                                    end=read.reference_end + polya_length - 1,
                                    strand=real_strand,
                                    name="polya"
                                )
                            )
                        elif real_strand == "-":
                            features_list.append(
                                GenomicLoci(
                                    chromosome=self.region.chromosome,
                                    start=read.reference_start - polya_length + 1,
                                    end=read.reference_start,
                                    strand=real_strand,
                                    name="polya"
                                )
                            )
                        else:
                            pass

                read = Reads(
                    chromosome=self.region.chromosome,
                    start=min(map(lambda x: x.start, exon_bound + features_list)),
                    end=max(map(lambda x: x.end, exon_bound + features_list)),
                    strand=self.region.strand,
                    id=read.query_name,
                    exons=exon_bound,
                    introns=intron_bound,
                    polya_length=polya_length,
                    m6a=m6a_loci,
                    features=features_list
                )

                if read.start < self.region.start or read.end > self.region.end:
                    continue

                self.data.append(read)

        except IOError as err:
            logger.error('There is no .bam file at {0}'.format(self.path))
            logger.error(err)
        except ValueError as err:
            logger.error(self.path)
            logger.error(err)

    def load(
            self,
            region: GenomicLoci,
            *args,
            **kwargs):
        u"""
        loading data
        :param region: the plotting region
        :return:
        """

        self.region = region

        # load from bed file
        if self.is_bed:
            self.load_bed()
        else:
            self.load_bam()

        tmp_df = pd.DataFrame(
            map(lambda x: x.to_dict(), self.data)
        )
        tmp_df["list_index"] = range(len(self.data))
        if self.exon_focus:
            e_f_use = []
            for read in self.data:
                tmp_ind = []
                for f_e in self.exon_focus:
                    tmp_ind.append("1" if f_e in str(read) else "0")
                e_f_use.append("_".join(tmp_ind))
            tmp_df["exon_group"] = e_f_use
        else:
            tmp_df["exon_group"] = "0"
        self.meta = self.df_sort(tmp_df)

    def len(self, scale: Union[int, float] = 0.005) -> int:
        u"""
        the length of reference to draw in final plots, default using the quarter of number of transcripts
        """
        size = len(self.data)
        return int(max(size * scale, 1))


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
u"""
This script contains the functions to draw different images
"""
import math
from collections import defaultdict
from copy import deepcopy
from typing import Dict, List, Optional, Union

import matplotlib as mpl
import numpy as np
import seaborn as sns

from adjustText import adjust_text
from loguru import logger
from matplotlib import pylab
from matplotlib.font_manager import FontProperties
from matplotlib.patches import PathPatch
from matplotlib.path import Path
from matplotlib.textpath import TextPath
from matplotlib.transforms import Affine2D
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from scipy.cluster.hierarchy import dendrogram, linkage
from scipy.stats import gaussian_kde, zscore

from sashimi.anno.theme import Theme
from sashimi.base.GenomicLoci import GenomicLoci
from sashimi.base.ReadDepth import ReadDepth
from sashimi.base.Stroke import Stroke
from sashimi.conf.config import CLUSTERING_METHOD, DISTANCE_METRIC
from sashimi.file.File import File
from sashimi.file.HiCMatrixTrack import HiCTrack
from sashimi.file.ReadSegments import ReadSegment
from sashimi.file.Reference import Reference


def get_limited_index(num, length):
    u"""
    Created by Zhang yiming at 2018.12.19
    Due to the original author didn't draw any element out of provided range
    So the scripts will through a lot of IndexError
    This function is used to scale that index into the reasonable range
    :param num: current index
    :param length: the list or numpy array length
    :return: (int, bool), 0 <= num <= length - 1, and modified or not
    """
    if num < 0:
        return 0, True

    if num >= length:
        return length - 1, True

    return num, False


def cubic_bezier(pts, t):
    """
    Get points in a cubic bezier.
    """
    p0, p1, p2, p3 = pts
    p0 = np.array(p0)
    p1 = np.array(p1)
    p2 = np.array(p2)
    p3 = np.array(p3)
    return p0 * (1 - t) ** 3 + 3 * t * p1 * (1 - t) ** 2 + 3 * t ** 2 * (1 - t) * p2 + t ** 3 * p3


def __merge_exons__(exons: List[List[int]]):
    u"""
    merge the overlap exons into one
    :param exons: sorted list of start and end sites of exons [[start, end], [start, end]]
    """
    exons = deepcopy(exons)
    res = []
    i, exon = 1, exons[0]
    while i < len(exons):
        curr_exon = exons[i]

        if curr_exon[0] <= exon[1]:
            exon[1] = curr_exon[1]
        else:
            res.append(exon)
            exon = curr_exon
        i += 1

    res.append(exon)
    return res


def init_graph_coords(region: GenomicLoci, exons: Optional[List[List[int]]] = None, exon_scale=1,
                      intron_scale=.5) -> np.array:
    u"""
    init the default
    :param region: the plot region
    :param exons: list of start and end sites of exons [[start, end], [start, end]]
    :param exon_scale: the scale of exon, default set to 1
    :param intron_scale: the scale of intron, default set to 0.5 -> the intron showed in final will be half size of real
    """
    graph_coords = np.zeros(len(region), dtype=int)

    if exons:
        # if there have exons, init graph_coords by exon and intron scales
        for i in range(0, exons[0][0] - region.start):
            graph_coords[i] = (i - 0) * intron_scale

        exons = __merge_exons__(exons)
        for i in range(0, len(exons)):
            exon = exons[i]

            if i > 0:
                intron = [exons[i - 1][1], exons[i][0]]

                for j in range(intron[0], intron[1]):
                    graph_coords[j - region.start] = graph_coords[intron[0] - region.start - 1] + (
                            j - intron[0] + 1) * intron_scale

            for j in range(exon[0], exon[1] + 1):
                graph_coords[j - region.start] = graph_coords[exon[0] - region.start - 1] + (
                        j - exon[0] + 1) * exon_scale

        intron = [exons[-1][-1], region.end]
        for i in range(intron[0], intron[1]):
            graph_coords[i - region.start] = graph_coords[intron[0] - region.start - 1] + (
                    i - intron[0] + 1) * intron_scale
    else:
        # if there is not any exons, just init graph_coords by region
        for i, j in enumerate(range(region.start, region.end + 1)):
            graph_coords[j - region.start] = i

    # if there is no exon or intron in last of region
    if graph_coords[-1] == 0:
        current_max = max(np.where(graph_coords == np.max(graph_coords))[-1])
        for i in np.where(graph_coords == 0)[0]:
            if i > current_max:
                graph_coords[i] = max(graph_coords) + 1
    return graph_coords


def set_x_ticks(
        ax: mpl.axes.Axes,
        region: GenomicLoci,
        graph_coords: Optional[Union[Dict, np.ndarray]] = None,
        sequence: Optional[Dict[int, str]] = None,
        log_trans: Optional[str] = None,
        nx_ticks: int = 4, font_size: int = 6, **kwargs):
    Theme.set_theme(ax, "blank")
    if graph_coords is None:
        graph_coords = init_graph_coords(region)

    # @2018.12.19 unnecessary text in figure
    x_label = 'Genomic coordinate (%s), "%s" strand' % (region.chromosome, region.strand)

    if log_trans:
        if log_trans in ["2", "10"]:
            log_trans = f"log{log_trans}"
        x_label = f"{x_label}, y axis is {log_trans} transformed"

    ax.hlines(y=0, xmin=0, xmax=max(graph_coords), color="black", lw=1)
    ax.text(x=graph_coords[len(graph_coords) // 2], y=-2.8, s=x_label, fontsize=font_size, ha="center", va="top")

    bk = 1
    if not sequence and nx_ticks > 1:
        bk = len(graph_coords) // (nx_ticks - 1)
    line_space = {}
    for i in range(0, len(graph_coords), bk):
        line_space[graph_coords[i]] = i + region.start
    line_space[max(graph_coords)] = region.end

    if sequence:
        for i, seq in sequence.items():
            relative_i = graph_coords[i - region.start]
            if relative_i in line_space.keys():
                temp_txs = "{}\n{}".format(sequence.get(i, ""), line_space[relative_i])
            else:
                temp_txs = "\n{}".format(sequence.get(i, ""))
            line_space[relative_i] = temp_txs

    for x, s in line_space.items():
        ax.vlines(x=x, ymin=-.5, ymax=0, color="black", lw=1)
        ax.text(x=x, y=-1, s=s, fontsize=font_size, ha="center", va="top")

    ax.set_ylim(-3.5, 1)
    ax.set_xlim(0, max(graph_coords))


def set_y_ticks(
        ax: mpl.axes.Axes,
        label: str,
        graph_coords: Union[Dict, np.array],
        max_used_y_val: Union[int, float],
        min_used_y_val: Optional[Union[int, float]] = None,
        distance_between_label_axis: float = .1,
        n_y_ticks: int = 4,
        theme: str = "ticks",
        font_size: int = 5,
        show_y_label: bool = True,
        set_label_only: bool = False,
        y_axis_skip_zero: bool = True,
        **kwargs
):
    u"""
    The y ticks are formatted here
    @2019.03.31 add little check here to make sure the y-axis shows the real value
    """
    # set y ticks, y label and label
    Theme.set_theme(ax, theme)

    if min_used_y_val is None:
        min_used_y_val, _ = ax.get_ylim()

    if not set_label_only:
        ax.set_xlim(0, max(graph_coords))

        max_ = max_used_y_val
        plus = 0.2
        while max_ > 10:
            max_ /= 10
            plus /= 10

        plus = (max_used_y_val - min_used_y_val) * plus
        ax.set_ylim(min_used_y_val - plus, plus + max_used_y_val)
        ax.spines["left"].set_bounds(min_used_y_val, max_used_y_val)

        universal_y_ticks = pylab.linspace(min_used_y_val, max_used_y_val, n_y_ticks + 1)
        # add zero into strand-aware plot
        universal_y_ticks = np.unique(np.append(universal_y_ticks, 0))
        universal_y_ticks = sorted(universal_y_ticks)
        curr_y_tick_labels = []

        for lab in universal_y_ticks:
            if y_axis_skip_zero and abs(lab - 0) < 0.0000000001:
                # Exclude label for 0
                curr_y_tick_labels.append("")
            else:
                curr_y_tick_labels.append("%.1f" % lab if lab % 1 != 0 else "%d" % lab)
        u"""
        @2019.01.04
        If there is no bam file, draw a blank y-axis 
        """
        ax.set_yticks(universal_y_ticks)
        ax.set_yticklabels(curr_y_tick_labels, fontsize=font_size)
        ax.yaxis.set_ticks_position('left')

    """
    Plot y labels
    @2018.12.20 using BAM label as y label
    @2019.01.04 change the standards of distance between y label and y-axis
    """
    if show_y_label:
        ax.text(
            x=-1 * distance_between_label_axis * max(graph_coords),
            y=(max_used_y_val + min_used_y_val) / 2,
            s=label, fontsize=font_size, ha="right"
        )


def set_focus(
        ax: mpl.axes.Axes,
        graph_coords: Union[Dict, np.array],
        focus: Dict[int, int]
):
    for left, right in focus.items():
        try:
            left, right = graph_coords[left], graph_coords[right]
            fill_x = [left, right, right, left]

            y1, y2 = ax.get_ylim()
            fill_y = [y1, y1, y2, y2]
            ax.fill(fill_x, fill_y, alpha=0.1, color='grey')
        except IndexError as err:
            logger.warning("focus region is out of bound: " + str(err))


def set_indicator_lines(
        ax: mpl.axes.Axes,
        graph_coords: Union[Dict, np.array],
        sites: Dict[int, str],
        min_y_used: Union[int, float] = 0,
        max_y_used: Union[int, float] = None
):
    if sites is None:
        return

    if not max_y_used:
        min_y_used, max_y_used = ax.get_ylim()

    for site, color in sites.items():
        try:
            ax.vlines(
                x=graph_coords[site],
                ymin=min_y_used,
                ymax=max_y_used,
                color=color,
                linestyles="dashed",
                lw=0.5
            )
        except IndexError as err:
            logger.warning("Indicator line is out of bound: " + str(err))


def plot_stroke(
        ax: mpl.axes.Axes,
        data: List[Stroke],
        graph_coords: Optional[Union[Dict, np.ndarray]] = None,
        font_size: int = 5,
        distance_between_label_axis: Union[int, float] = .1,
        theme: str = "blank",
        **kwargs
):
    u"""
    plot stoke
    """

    strokes = sorted(data, key=lambda x: [x.start, x.end])

    for i, stroke in enumerate(strokes):
        try:
            ax.hlines(
                y=i,
                xmin=graph_coords[stroke.start],
                xmax=graph_coords[stroke.end],
                color=stroke.color, lw=2)
            ax.text(
                -1 * distance_between_label_axis * max(graph_coords),
                i + .2,
                stroke.label,
                fontsize=font_size,
                color=stroke.color
            )
        except IndexError as err:
            logger.warning(f"stroke is out of bound: {err}")

    Theme.set_theme(ax, theme)
    ax.set_xlim(left=0, right=max(graph_coords))
    ax.set_ylim(bottom=-1, top=len(strokes))


def plot_reference(
        ax: mpl.axes.Axes,
        obj: Reference,
        graph_coords: Optional[Union[Dict, np.ndarray]] = None,
        font_size: int = 5,
        show_gene: bool = False,
        show_id: bool = False,
        transcripts: Optional[List[str]] = None,
        remove_empty_transcripts: bool = False,
        choose_primary: bool = False,
        color: Optional[str] = None,
        reverse_minus: bool = False,
        theme: str = "blank",
        y_loc: int = 0,
        exon_width: float = .3,
        plot_domain: bool = False,
        show_exon_id: bool = False,
        raster: bool = True,
        distance_between_label_axis: float = 0.3,
        **kwargs
):
    u"""
    Plot the structure of reference
    """
    Theme.set_theme(ax, theme)
    ax.set_xlim(0, max(graph_coords))

    region = obj.region

    data = sorted(obj.data)
    if not data:
        return

    if graph_coords is None:
        graph_coords = init_graph_coords(region)

    color = "k" if not color else color

    if choose_primary and (len(transcripts) == 0 or transcripts is None):
        transcripts = []
        # For each gene, you can choose only one transcript to plot.
        genes = defaultdict(list)
        for transcript in data:
            if transcript.category == 'interval':
                continue
            genes[transcript.gene_id].append(transcript)

        for _, transcripts_list in genes.items():
            primary_transcripts = sorted(
                transcripts_list, key=lambda i_: len(i_), reverse=True)[0]
            transcripts.append(primary_transcripts.transcript_id)
    elif choose_primary and (len(transcripts) != 0 or transcripts is not None):
        logger.warning(
            "--transcripts-to-show is prior to --choose-primary, and primary transcript won't be presented.")
    else:
        pass

    """
    @2018.12.26
    Maybe I'm too stupid for this, using 30% of total length of x axis as the gap between text with axis
    """

    for transcript in obj.data:
        # ignore the unwanted transcript
        if transcripts and not (set(transcripts) & set(transcript.ids())):
            continue

        # ignore transcripts without any exons
        if remove_empty_transcripts and not transcript.exons:
            continue

        strand = transcript.strand
        # @2018.12.20 add transcript id, based on fixed coordinates
        if transcript.transcript:
            if show_gene and transcript.gene and transcript.gene_id != transcript.transcript_id:
                if show_id:
                    ax.text(x=-1 * distance_between_label_axis * max(graph_coords), y=y_loc + 0.25,
                            s=transcript.gene_id, fontsize=font_size, ha="right")
                    ax.text(x=-1 * distance_between_label_axis * max(graph_coords), y=y_loc - 0.25,
                            s=transcript.transcript_id, fontsize=font_size, ha="right")
                else:
                    ax.text(x=-1 * distance_between_label_axis * max(graph_coords), y=y_loc,
                            s=transcript.gene + " | " + transcript.transcript,
                            fontsize=font_size, ha="right")
            else:
                ax.text(x=-1 * distance_between_label_axis * max(graph_coords), y=y_loc - 0.1, s=transcript.transcript,
                        fontsize=font_size, ha="right")

        # @2018.12.19
        # s and e is the start and end site of single exon
        # @2022.05.13
        # add index to avoid label overlapping of neighbor exon
        for ind, exon in enumerate(transcript.exons):
            s, e, strand = region.relative(
                exon.start), region.relative(exon.end), exon.strand
            x = [
                graph_coords[s], graph_coords[e],
                graph_coords[e], graph_coords[s]
            ]
            y = [
                y_loc - exon_width / 2, y_loc - exon_width / 2,
                y_loc + exon_width / 2, y_loc + exon_width / 2
            ]
            ax.fill(x, y, color, lw=.5, zorder=20)

            if show_exon_id:
                y_loc_offset = 0.1 if ind % 2 == 0 else - 0.2
                ax.text(x=(graph_coords[s] + graph_coords[s]) / 2, y=y_loc + y_loc_offset,
                        s=exon.name, fontsize=font_size / 2, ha="right")

        # @2018.12.21
        # change the intron range
        # Draw intron.
        # if transcript's category is interval, then don't plot the intron.
        if transcript.category != "interval":
            intron_sites = [
                graph_coords[region.relative(transcript.start)],
                graph_coords[region.relative(transcript.end)]
            ]
            ax.plot(intron_sites, [y_loc, y_loc], color=color, lw=0.5)

            # @2018.12.23 fix intron arrows issues
            # Draw intron arrows.
            max_ = graph_coords[region.relative(transcript.end)]
            min_ = graph_coords[region.relative(transcript.start)]
            length = max_ - min_
            narrows = math.ceil(length / max(graph_coords) * 50)

            if narrows > 0:
                spread = .2 * length / narrows

                for i in range(narrows):
                    loc = float(i) * length / narrows + graph_coords[region.relative(transcript.start)]
                    if strand == '+' or reverse_minus:
                        x = [loc - spread, loc, loc - spread]
                    else:
                        x = [loc + spread, loc, loc + spread]
                    y = [y_loc - exon_width / 5, y_loc, y_loc + exon_width / 5]
                    ax.plot(x, y, lw=.5, color=color, rasterized=raster)

        y_loc += 1  # if transcript.transcript else .5

        if plot_domain and obj.domain and transcript.transcript_id in obj.domain.pep:
            current_domains = obj.domain.pep[transcript.transcript_id]

            for sub_current_domain in current_domains:
                if not (sub_current_domain.start <= region.end and
                        sub_current_domain.end >= region.start):
                    continue

                for sub_exon in sub_current_domain.exons:

                    for exon in sub_exon:
                        s, e = region.relative(
                            exon.start), region.relative(exon.end)
                        if e <= 0 or s > len(region):
                            continue
                        s = 0 if s < 0 else s
                        e = len(region) - 1 if e > len(region) else e

                        x = [
                            graph_coords[s], graph_coords[e],
                            graph_coords[e], graph_coords[s]
                        ]
                        y = [
                            y_loc - exon_width / 4, y_loc - exon_width / 4,
                            y_loc + exon_width / 4, y_loc + exon_width / 4
                        ]
                        ax.fill(x, y, color, lw=.5,
                                zorder=20, rasterized=raster)

                    # @2022.05.13
                    intron_relative_s = region.relative(
                        min(map(lambda x_: x_.end, sub_exon)))
                    intron_relative_s = intron_relative_s if intron_relative_s >= 0 else 0
                    if intron_relative_s > len(region):
                        continue

                    intron_relative_e = region.relative(
                        max(map(lambda x_: x_.start, sub_exon)))
                    intron_relative_e = len(
                        region) - 1 if intron_relative_e > len(region) else intron_relative_e
                    if intron_relative_e <= 0:
                        continue

                    intron_sites = [graph_coords[intron_relative_s],
                                    graph_coords[intron_relative_e]]
                    if len(sub_exon) != 1:
                        ax.plot(intron_sites, [y_loc, y_loc],
                                color=color, lw=0.2, rasterized=raster)

                ax.text(x=-1, y=y_loc - 0.125, s=f"{sub_current_domain.gene}|{transcript.transcript_id}",
                        fontsize=font_size / 2, ha="right")

                y_loc += 0.25

        # offset for next term.
        y_loc += 0.75
    if obj.local_domain:
        for base_name, current_domain in obj.local_domain.items():
            for sub_current_domain in current_domain:
                if not (sub_current_domain.start <= region.end and
                        sub_current_domain.end >= region.start):
                    continue

                for sub_exon in sub_current_domain.exons:

                    for exon in sub_exon:
                        s, e = region.relative(
                            exon.start), region.relative(exon.end)
                        if e <= 0 or s > len(region):
                            continue
                        s = 0 if s < 0 else s
                        e = len(region) - 1 if e > len(region) else e

                        x = [
                            graph_coords[s], graph_coords[e],
                            graph_coords[e], graph_coords[s]
                        ]
                        y = [
                            y_loc - exon_width / 4, y_loc - exon_width / 4,
                            y_loc + exon_width / 4, y_loc + exon_width / 4
                        ]
                        ax.fill(x, y, color, lw=.5,
                                zorder=20, rasterized=raster)

                    # @2022.05.13
                    intron_relative_s = region.relative(
                        min(map(lambda x_: x_.end, sub_exon)))
                    intron_relative_s = intron_relative_s if intron_relative_s >= 0 else 0
                    if intron_relative_s > len(region):
                        continue

                    intron_relative_e = region.relative(
                        max(map(lambda x_: x_.start, sub_exon)))
                    intron_relative_e = len(
                        region) - 1 if intron_relative_e > len(region) else intron_relative_e
                    if intron_relative_e <= 0:
                        continue

                    intron_sites = [graph_coords[intron_relative_s],
                                    graph_coords[intron_relative_e]]
                    if len(sub_exon) != 1:
                        ax.plot(intron_sites, [y_loc, y_loc],
                                color=color, lw=0.2, rasterized=raster)

                ax.text(x=-1, y=y_loc - 0.125, s=f"{sub_current_domain.gene}|{base_name}",
                        fontsize=font_size / 2, ha="right")

            y_loc += 1

    # @2022.05.13 Set y lim using y_loc value.
    ax.set_ylim(-.5, y_loc + .5)


def plot_density(
        ax: mpl.axes.Axes,
        obj: Optional[File] = None,
        data: Optional[ReadDepth] = None,
        region: Optional[GenomicLoci] = None,
        graph_coords: Optional[Union[Dict, np.ndarray]] = None,
        color="blue",
        font_size: int = 8,
        show_junction_number: bool = True,
        junction_number_font_size: int = 5,
        n_y_ticks: int = 4,
        distance_between_label_axis: float = .1,
        show_y_label: bool = True,
        y_label: str = "",
        theme: str = "ticks_blank",
        max_used_y_val: Optional[float] = None,
        raster: bool = False,
        **kwargs
):
    u"""
    draw density plot
    :param ax: mpl.axes.Axes
    :param data: File
    :param data: ReadDepth
    :param region: GenomicLoci
    :param graph_coords:
    :param font_size: the font size for ticks, y-axis label and title
    :param show_junction_number: whether to show the number of junctions
    :param distance_between_label_axis: distance between y-axis label and y-axis ticks
    :param n_y_ticks: number of y ticks
    :param junction_number_font_size:
    :param obj: Bam or Bigwig object
    :param color: color for this density plot
    :param show_y_label: whether to show y-axis label
    :param y_label: the text of y-axis title
    :param theme: the theme name
    :param max_used_y_val: used to set same max y-axis
    :param raster:
    :param kwargs:
    :return:
    """
    if obj:
        assert obj.region is not None, "please load data first"
        region = obj.region
        if graph_coords is None:
            graph_coords = init_graph_coords(region)

        if not y_label:
            y_label = obj.label

    elif not (region and data):
        raise ValueError("please input obj or region and data")

    if data is None:
        data = obj.data
    jxns = data.junctions_dict

    if max_used_y_val is None:
        max_used_y_val = max(data.plus)
        if max_used_y_val % 2 == 1:
            max_used_y_val += 1

    min_used_y_val = -1 * max(data.minus) if data.minus is not None else 0

    # Reduce memory footprint by using incremented graph_coords.
    x, y1, y2 = [], [], []
    for i in range(len(graph_coords)):
        x.append(graph_coords[i])
        y1.append(data.plus[i] if data.plus is not None else 0)
        y2.append(-data.minus[i] if data.minus is not None else 0)

    ax.fill_between(x, y1, y2=y2, color=color, lw=0, step="post", rasterized=raster)

    if jxns:
        # sort the junctions by intron length for better plotting look
        jxns_sorted_list = sorted(jxns.keys(), key=lambda x: (x.end - x.start, x.start, x.end), reverse=True)

        if not jxns:
            max_junction_count, min_junction_count = 0, 0
        else:
            max_junction_count = max(jxns.values())
            min_junction_count = min(jxns.values())
        junction_count_gap = max_junction_count - min_junction_count

        jxn_numbers = []
        for plotted_count, jxn in enumerate(jxns_sorted_list):
            leftss, rightss = jxn.start, jxn.end

            # @2022.09.26
            # Skip these too short span junction for avoiding plotting junction number
            overlap_length = min(rightss, region.end) - max(leftss, region.start) + 1
            if not overlap_length / len(region) > 0.5 and not overlap_length / len(jxns) > 0.5:
                continue

            # @2018.12.19
            # set junctions coordinate here
            # the junction out of boundaries, set the boundaries as coordinate
            ss1_idx, ss1_modified = get_limited_index(leftss - region.start, len(graph_coords))
            ss2_idx, ss2_modified = get_limited_index(rightss - region.start, len(graph_coords))
            u"""
            @2019.01.14
            add two new variables to make it clear which one is index, which one is genomic site 
            """
            ss1, ss2 = graph_coords[ss1_idx], graph_coords[ss2_idx]
            # draw junction on bottom
            if plotted_count % 2 == 0:
                current_height = abs(3 * min_used_y_val / 4)
                left_dens, right_dens = abs(data.curr_min(ss1_idx)), abs(data.curr_min(ss2_idx))
                pts = [
                    (ss1, -left_dens if not ss1_modified else -left_dens - current_height),
                    (ss1, -current_height),
                    (ss2, -current_height),
                    (ss2, -right_dens if not ss2_modified else -right_dens - current_height)
                ]
            # draw junction on top
            else:
                current_height = abs(3 * max_used_y_val / 4)
                left_dens, right_dens = data.curr_max(ss1_idx), data.curr_max(ss2_idx)
                pts = [
                    (ss1, left_dens if not ss1_modified else left_dens + current_height),
                    (ss1, left_dens + current_height),
                    (ss2, right_dens + current_height),
                    (ss2, right_dens if not ss2_modified else right_dens + current_height)
                ]

            if show_junction_number:
                midpt = cubic_bezier(pts, .5)

                t = ax.text(
                    midpt[0], midpt[1],
                    '{0}'.format(round(jxns[jxn], 2)),
                    fontsize=junction_number_font_size,
                    ha='center', va='center',
                    backgroundcolor='w'
                )

                # @2018.12.19 transparent background
                t.set_bbox(dict(alpha=0))
                jxn_numbers.append(t)

            a = Path(pts, [Path.MOVETO, Path.CURVE4, Path.CURVE4, Path.CURVE4])

            """
            @2018.12.26
            scale the junctions line width
            """
            if junction_count_gap > 0:
                line_width = (jxns[jxn] - min_junction_count) / junction_count_gap
            else:
                line_width = 0

            ax.add_patch(PathPatch(a, ec=color, lw=line_width + 0.2, fc='none'))

        adjust_text(jxn_numbers, force_text=0.2, arrowprops=dict(arrowstyle="-", color='black', lw=1), autoalign="y")

    if obj and obj.title:
        ax.text(max(graph_coords) - len(obj.title), max_used_y_val, obj.title, color=color, fontsize=font_size)

    if data.strand_aware:
        max_used_y_val = max(abs(min_used_y_val), max_used_y_val)
        min_used_y_val = -max(abs(min_used_y_val), max_used_y_val) if data.minus is not None else 0

    set_y_ticks(
        ax, label=y_label, theme=theme,
        graph_coords=graph_coords,
        max_used_y_val=max_used_y_val,
        min_used_y_val=min_used_y_val,
        n_y_ticks=n_y_ticks,
        distance_between_label_axis=distance_between_label_axis,
        font_size=font_size,
        show_y_label=show_y_label,
        y_axis_skip_zero=False if data.strand_aware else True
    )


def plot_site_plot(
        ax: mpl.axes.Axes,
        obj: File,
        graph_coords: Optional[Union[Dict, np.ndarray]] = None,
        color="blue",
        font_size: int = 8,
        n_y_ticks: int = 3,
        distance_between_label_axis: float = .1,
        show_y_label: bool = True,
        y_label: str = "",
        strand_choice: str = None,
        theme="ticks",
        raster: bool = False,
        **kwargs
):
    """
    :param ax:
    :param obj:
    :param graph_coords:
    :param color:
    :param font_size:
    :param n_y_ticks:
    :param distance_between_label_axis:
    :param show_y_label:
    :param y_label:
    :param strand_choice:
    :param theme:
    :param raster:
    :return:
    """
    region = obj.region
    if graph_coords is None:
        graph_coords = init_graph_coords(region)

    if not y_label:
        y_label = obj.label

    if not show_y_label:
        y_label = ""

    plus, minus = obj.data.site_plus, obj.data.site_minus
    max_height, min_height = max(plus), min(minus)
    max_val = max(max_height, abs(min_height))

    for label, array_plot in zip(['plus', 'minus'], [plus, minus]):
        if strand_choice != "all" and label != strand_choice:
            continue

        array_hist = np.repeat(graph_coords, np.abs(array_plot).astype(np.int))
        try:
            kde = gaussian_kde(array_hist)
            fit_value = kde.pdf(graph_coords)
        except (ValueError, np.linalg.LinAlgError):
            # logger.warning(err)
            # logger.warning(traceback.format_exc())
            continue

        fit_value = fit_value / fit_value.max()
        ax.bar(graph_coords, array_plot, color=color, rasterized=raster)
        ax.plot(graph_coords,
                fit_value * array_plot.max() if label == 'plus' else fit_value * array_plot.min(),
                c=color, lw=1)

    # set the y limit
    # set y ticks, y label and label
    set_y_ticks(
        ax, label=y_label, theme=theme,
        graph_coords=graph_coords,
        max_used_y_val=1.1 * max_val,
        min_used_y_val=-1.1 * max_val,
        n_y_ticks=n_y_ticks,
        font_size=font_size,
        show_y_label=False,
        distance_between_label_axis=distance_between_label_axis,
        y_axis_skip_zero=False
    )


def plot_heatmap(
        ax: mpl.axes.Axes,
        cbar_ax: mpl.axes.Axes,
        data: Dict[str, ReadDepth],
        graph_coords: Optional[Union[Dict, np.ndarray]] = None,
        color="viridis",
        font_size: int = 8,
        distance_between_label_axis: float = .1,
        show_y_label: bool = True,
        theme: str = "ticks",
        y_label: str = "",
        do_scale: bool = False,
        clustering: bool = False,
        clustering_method: str = "ward",
        distance_metric: str = "euclidean",
        raster: bool = True,
        show_row_names: bool = False,
        vmin=None, vmax=None,
        **kwargs
):
    u"""

    :param theme:
    :param graph_coords:
    :param ax: whether to scale the matrix
    :param cbar_ax: whether to scale the matrix
    :param font_size:
    :param data:
    :param distance_between_label_axis:
    :param show_y_label:
    :param y_label:
    :param do_scale: whether to scale the matrix
    :param clustering: whether reorder matrix by clustering
    :param clustering_method: same as  scipy.cluster.hierarchy.linkage
    :param distance_metric: same as scipy.spatial.distance.pdist
    :param color: used for seaborn.heatmap, see: https://matplotlib.org/3.5.1/tutorials/colors/colormaps.html
                'binary', 'gist_yarg', 'gist_gray', 'gray', 'bone',
                'pink', 'spring', 'summer', 'autumn', 'winter', 'cool',
                'Wistia', 'hot', 'afmhot', 'gist_heat', 'copper'
    :param raster: whether to draw image in raster mode
    :param show_row_names:
    :param vmin: Values to anchor the colormap, otherwise they are inferred from the data and other keyword arguments.
    :param vmax: Values to anchor the colormap, otherwise they are inferred from the data and other keyword arguments.
    """
    labels = list(data.keys())
    mtx = np.array([x.wiggle for x in data.values()])

    if clustering and len(mtx) > 1:
        assert clustering_method in CLUSTERING_METHOD, f"clustering_method {clustering_method} is not supported."
        assert distance_metric in DISTANCE_METRIC, f"distance_metric {distance_metric} is not supported"

        order = dendrogram(
            linkage(mtx, method=clustering_method, metric=distance_metric),
            orientation='right'
        )

        mtx = mtx[order["leaves"], :]
        labels = [labels[x] for x in order["leaves"]]

    if do_scale:
        """
        y = (x – mean) / standard_deviation
        """
        mtx = zscore(mtx, axis=1)

    if not show_row_names:
        labels = False

    sns.heatmap(mtx, ax=ax, cmap=color, cbar_ax=cbar_ax,
                xticklabels=False, yticklabels=labels,
                center=False, rasterized=raster, vmin=vmin, vmax=vmax)

    ax.tick_params(axis='both', which='major', labelsize=font_size, rotation=0)
    cbar_ax.tick_params(labelsize=font_size)

    ymax, ymin = ax.get_ylim()
    set_y_ticks(
        ax, label=y_label, theme=theme,
        graph_coords=graph_coords,
        max_used_y_val=ymax, min_used_y_val=ymin,
        distance_between_label_axis=distance_between_label_axis,
        font_size=font_size,
        show_y_label=show_y_label,
        set_label_only=True
    )


def plot_hic(
        ax: mpl.axes.Axes,
        cbar_ax: mpl.axes.Axes,
        obj: List[HiCTrack],
        show_legend: bool = True,
        graph_coords: Optional[Union[Dict, np.ndarray]] = None,
        color: str = "RdYlBu_r",
        font_size: int = 8,
        distance_between_label_axis: float = .1,
        show_y_label: bool = True,
        theme: str = "ticks",
        y_label: str = "",
        n_y_ticks: int = 4,
        raster: bool = True,
        **kwargs
):
    assert len(obj) == 1, "HiC plot only support one file"

    obj = obj[0]

    if graph_coords is None:
        graph_coords = init_graph_coords(
            obj.region
        )

    if not y_label:
        y_label = obj.label

    ax.pcolormesh(obj.x - obj.region.start, obj.y, np.flipud(obj.matrix),
                  cmap=color, rasterized=raster)
    ax.set_xlim(0, len(obj.region))
    ax.set_ylim(0, obj.matrix.shape[1])

    if show_legend:
        cbar = pylab.colorbar(
            mappable=mpl.cm.ScalarMappable(
                norm=mpl.colors.Normalize(
                    vmin=0, vmax=np.percentile(obj.matrix.diagonal(1), 80)),
                cmap=color),
            cax=cbar_ax)
        cbar.ax.tick_params(labelsize=font_size)
        if obj.trans:
            legend_ticks = cbar.get_ticks().tolist()
            legend_ticks[0] = f"{legend_ticks[0]}\n{obj.trans}"
            cbar.set_ticklabels(legend_ticks)

    ax.axis("off")
    set_y_ticks(
        ax, label=y_label, theme=theme,
        graph_coords=graph_coords,
        max_used_y_val=obj.depth,
        distance_between_label_axis=distance_between_label_axis,
        font_size=font_size,
        show_y_label=show_y_label,
        n_y_ticks=n_y_ticks
    )


def plot_line(
        ax: mpl.axes.Axes,
        data: Dict[str, ReadDepth],
        graph_coords: Union[Dict, np.ndarray],
        font_size: int = 8,
        distance_between_label_axis: float = .1,
        show_y_label: bool = True,
        y_label: str = "",
        line_attrs: Optional[Dict[str, Dict]] = None,
        theme: str = "ticks_blank",
        n_y_ticks: int = 4,
        show_legend: bool = False,
        legend_position: str = "upper right",
        legend_ncol: int = 0,
        max_used_y_val: Optional[float] = None,
        **kwargs
):
    u"""

    :param ax: whether to scale the matrix
    :param font_size:
    :param data:
    :param distance_between_label_axis:
    :param show_y_label:
    :param graph_coords:
    :param y_label:
    :param line_attrs: dict contains attributes like color and type for each line
    :param theme: the theme name
    :param n_y_ticks: the number of y ticks
    :param show_legend: whether to show legend
    :param legend_position:
    :param legend_ncol:
    :param max_used_y_val:
    """
    max_y_val = 0
    for ylab, val in data.items():
        attr = line_attrs.get(ylab, {}) if line_attrs else {}

        x, y = [], []
        for i, j in enumerate(val.wiggle):
            x.append(graph_coords[i])
            y.append(j)
        ax.plot(x, y, label=ylab if show_y_label else "", **attr)

        max_y_val = max(max_y_val, max(val.wiggle))

    if show_legend:
        ax.legend(loc=legend_position,
                  ncol=int(len(data) / 1.5) if legend_ncol <= 0 else legend_ncol,
                  fancybox=False, shadow=False)

    set_y_ticks(
        ax, label=y_label, theme=theme,
        graph_coords=graph_coords,
        max_used_y_val=max_y_val if max_used_y_val is None else max_used_y_val,
        distance_between_label_axis=distance_between_label_axis,
        font_size=font_size,
        show_y_label=show_y_label,
        n_y_ticks=n_y_ticks, **kwargs
    )


def plot_igv_like(
        ax: mpl.axes.Axes,
        obj: Dict[str, ReadSegment],
        graph_coords: Optional[Union[Dict, np.ndarray]] = None,
        y_label: str = "",
        exon_color: Optional[str] = None,
        intron_color: Optional[str] = None,
        feature_color: Optional[str] = None,
        exon_width: float = .3,
        font_size: int = 8,
        n_y_ticks: int = 1,
        distance_between_label_axis: float = .1,
        show_y_label: bool = True,
        theme: str = "ticks_blank",
        raster: bool = False,
        **kwargs
):
    u"""

    :param show_y_label:
    :param n_y_ticks:
    :param feature_color:
    :param distance_between_label_axis:
    :param ax:
    :param obj:
    :param graph_coords:
    :param y_label:
    :param exon_color:
    :param intron_color:
    :param exon_width:
    :param font_size:
    :param theme:
    :param raster:
    :return:
    """

    assert len(obj) == 1, "IGV-like plot only support one file"

    obj = list(obj.values())[0]
    assert obj.region is not None, "please load data first"

    region = obj.region
    if graph_coords is None:
        graph_coords = init_graph_coords(region)

    if not y_label:
        y_label = obj.label
    y_loc = 0.5

    # Add this to skip zero coverage regions.
    if obj.meta is not None:

        for c_ind_list in obj.get_index():
            add_plot = False
            for c_ind in c_ind_list:
                c_data = obj.data[c_ind]
                # skip truncated reads in the given region
                if c_data.start < region.start or c_data.end > region.end:
                    continue

                for exon in c_data.exons:
                    s, e, strand = region.relative(
                        exon.start), region.relative(exon.end), exon.strand
                    if e < 0 or s > len(region):
                        continue

                    s = 0 if s < 0 else s
                    e = len(region) - 1 if e >= len(region) else e

                    x = [
                        graph_coords[s], graph_coords[e],
                        graph_coords[e], graph_coords[s]
                    ]
                    y = [
                        y_loc - exon_width, y_loc - exon_width,
                        y_loc + exon_width, y_loc + exon_width
                    ]

                    ax.fill(x, y, 'k' if not exon_color else exon_color,
                            lw=.5, zorder=20)
                    add_plot = add_plot | True

                for intron in c_data.introns:
                    s, e, strand = region.relative(
                        intron.start), region.relative(intron.end), intron.strand

                    if e < 0 or s > len(region):
                        continue
                    s = 0 if s < 0 else s
                    e = len(region) - 1 if e >= len(region) else e
                    intron_sites = [
                        graph_coords[s],
                        graph_coords[e]
                    ]
                    ax.plot(intron_sites, [y_loc, y_loc],
                            color="#4d4d4d" if not intron_color else intron_color,
                            lw=0.2, rasterized=raster)

                for feature in c_data.features:
                    if feature.start == -1:
                        continue

                    s, e, strand = region.relative(feature.start), region.relative(
                        feature.end), feature.strand
                    is_site = False

                    if s == e:
                        s = s - 1
                        is_site = True

                    if e < 0 or s > len(region):
                        continue

                    s = 0 if s < 0 else s
                    e = len(region) - 1 if e >= len(region) else e

                    x = [
                        graph_coords[s], graph_coords[e],
                        graph_coords[e], graph_coords[s]
                    ]
                    width_ratio = 0.75 if not is_site else 1.5

                    y = [
                        y_loc - exon_width * width_ratio, y_loc - exon_width * width_ratio,
                        y_loc + exon_width * width_ratio, y_loc + exon_width * width_ratio
                    ]
                    ax.fill(
                        x, y, 'r' if not feature_color else feature_color, lw=.2, zorder=20)
                    if is_site:
                        ax.scatter(graph_coords[s],
                                   y_loc + exon_width * width_ratio,
                                   c='b' if not feature_color else feature_color,
                                   s=.5,
                                   linewidths=(0,),
                                   rasterized=raster
                                   )
            if add_plot:
                y_loc += 1

    set_y_ticks(
        ax, label=y_label, theme=theme,
        graph_coords=graph_coords,
        max_used_y_val=y_loc,
        n_y_ticks=n_y_ticks,
        distance_between_label_axis=distance_between_label_axis,
        font_size=font_size,
        show_y_label=show_y_label,
    )


def plot_links(ax: mpl.axes.Axes,
               data: List[Stroke],
               graph_coords: Optional[Union[Dict, np.ndarray]] = None,
               max_y: int = -10, **kwargs):
    for stroke in sorted(data):
        leftss, rightss = graph_coords[stroke.start], graph_coords[stroke.end]

        step = (rightss - leftss) / 4

        pts = [
            (leftss, 0),
            (leftss + step, max_y),
            (rightss - step, max_y),
            (rightss, 0)
        ]
        a = Path(pts, [Path.MOVETO, Path.CURVE4, Path.CURVE4, Path.CURVE4])
        ax.add_patch(PathPatch(a, fc="none", ec=stroke.color, lw=1))

    ax.set_xlim(0, max(graph_coords))
    ax.set_ylim(max_y, 0)
    ax.axis("off")


def make_text_elements(text, x=0.0, y=0.0, width=1.0, height=1.0,
                       color='blue', edgecolor="black",
                       font=FontProperties(family='monospace')):
    tp = TextPath((0.0, 0.0), text, size=1, prop=font)
    bbox = tp.get_extents()
    bwidth = bbox.x1 - bbox.x0
    bheight = bbox.y1 - bbox.y0
    trafo = Affine2D()
    trafo.translate(-bbox.x0, -bbox.y0)
    trafo.scale(1 / bwidth * width, 1 / bheight * height)
    trafo.translate(x, y)
    tp = tp.transformed(trafo)
    return PathPatch(tp, facecolor=color, edgecolor=edgecolor)


def plot_motif(ax: mpl.axes.Axes,
               obj,  # list of weighted text
               graph_coords: Optional[Union[Dict, np.ndarray]] = None,
               width: float = 0.8,
               colors=None,
               width_per_character: float = 3.5,
               theme: str = "blank",
               **kwargs):
    Theme.set_theme(ax, theme)
    data = obj.data
    if colors is None:
        colors = ['#008000', '#cc0000', '#0000cc', '#ffb300']
        colors = {x: y for x, y in zip(["A", "T", "C", "G"], colors)}

    region = obj.region
    if graph_coords is None:
        graph_coords = init_graph_coords(region)

    # 在原始坐标轴上画motif
    ymin, ymax, xmin, xmax = 0, 0, \
                             graph_coords[min(data.keys()) - region.start], \
                             graph_coords[max(data.keys()) - region.start] + (1 + width) / 2

    # 在放大区画motif
    axin_width = width_per_character * (xmax - xmin) / len(graph_coords)
    bbox_to_left = (xmax * 2 - xmin) / len(graph_coords)
    draw_left = False
    if axin_width + bbox_to_left > 1:
        bbox_to_left = max(xmin / len(graph_coords) - axin_width - .1, 0)
        draw_left = True

    axins = inset_axes(ax, width=f"{axin_width * 100}%", height="100%",
                       loc='center left',
                       bbox_to_anchor=(bbox_to_left, 0, 1, 1),
                       bbox_transform=ax.transAxes)

    start_site = min(list(data.keys()))
    site = 0
    for idx, vals in data.items():
        site = idx - start_site
        init_height_pos, init_height_neg = 0, 0
        for text, height in vals.items():
            text_shape = make_text_elements(text,
                                            x=site + (1 - width) / 2,
                                            y=init_height_neg if height < 0 else init_height_pos,
                                            width=width, height=height,
                                            color=colors.get(text, "blue"),
                                            edgecolor=colors.get(text, "blue"))

            if height > 0:
                init_height_pos += height
            else:
                init_height_neg += height
            ymin, ymax = min(ymin, init_height_neg), max(ymax, init_height_pos)
            axins.add_patch(text_shape)
    axins.set_xlim(0, site + (1 + width) / 2)
    axins.set_ylim(ymin, ymax)
    axins.tick_params(bottom=False, top=False, left=False, right=False)
    axins.set_xticklabels([])
    axins.set_yticklabels([])

    # draw box
    y_height = ymax - ymin
    y_center = (ymax + ymin) / 2
    y_scale = .25
    ax.plot(
        [xmin, xmax, xmax, xmin, xmin],
        [ymin, ymin, ymin + y_height * y_scale, ymin + y_height * y_scale, ymin],
        "black")

    ax.set_xlim(min(graph_coords), max(graph_coords))
    ax.set_ylim(ymin, ymax)

    # 画两条线
    if draw_left:
        pos = len(graph_coords) * ((bbox_to_left + axin_width) * 100 + 5) / 100
        x = xmin
    else:
        pos = len(graph_coords) * bbox_to_left
        x = xmax
    ax.arrow(x, ymin + y_height * y_scale / 2,
             pos - x, y_center - (ymin + y_height * y_scale),
             head_width=1, fc='k', ec='k', rasterized=True)


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
u"""
Created by ygidtu@gmail.com at 2019.01.04

This script contains all the basic data types used by this suite of scripts

For better organization

Changelog:
    1. add relative to convert sites to relative coord
"""


class GenomicLoci(object):
    u"""
    Created by ygidtu at 2018.12.19

    A base class to handle the position relationships
    """

    __slots__ = [
        "chromosome",
        "start",
        "end",
        "strand",
        "gtf_line",
        "name"
    ]

    def __init__(self, chromosome, start, end, strand, name="", gtf_line=None):
        u"""
        init this class
        :param chromosome: str
        :param start: int
        :param end: int
        :param strand: strand information
        :param name: name of given feature
        :param strand: str
        """

        self.chromosome = chromosome

        self.start = int(start)
        self.end = int(end)
        self.gtf_line = gtf_line
        self.name = name

        if self.end < self.start:
            raise ValueError(f"End site should bigger than start site, not {self.start} -> {self.end}")
        if strand == ".":
            strand = "*"
        if strand not in ("+", "-", "*"):
            raise ValueError(f"strand should be + or -, not {strand}")

        self.strand = strand

    def __str__(self):
        u"""
        convert this to string
        :return:
        """
        return f"{self.chromosome}:{self.start}-{self.end}:{self.strand}"

    def __gt__(self, other):
        u"""
        if other downstream of other

        Note:
            make sure the wider range is upstream of narrower

            due to the sort of gtf file, therefore the transcript will be ahead of exons
        :param other:
        :return:
        """
        if self.chromosome != other.chromosome:
            return self.chromosome > other.chromosome

        if self.start != other.start:
            return self.start > other.start

        return self.end < other.end

    def __lt__(self, other):
        u"""
        if other is upstream of other

        Note:
            make sure the wider range is downstream of narrower

            due to the sort of gtf file, therefore the transcript will be ahead of exons
        :param other:
        :return:
        """
        if self.chromosome != other.chromosome:
            return self.chromosome < other.chromosome

        if self.start != other.start:
            return self.start < other.start

        return self.end > other.end

    def __eq__(self, other):
        u"""
        if two objects are the same
        :param other:
        :return:
        """
        return hash(self) == hash(other)

    def __add__(self, other):
        u"""
        merge two sites into one
        :param other:
        :return:
        """
        return GenomicLoci(
            chromosome=self.chromosome,
            start=min(self.start, other.start),
            end=max(self.end, other.end),
            strand=self.strand
        )

    def __hash__(self):
        u"""
        generate hash
        :return:
        """
        return hash((self.chromosome, self.start, self.end))

    def __len__(self) -> int:
        return self.end - self.start + 1

    def is_overlap(self, other):
        u"""
        whether two loci have any overlaps
        :param other: another GenomicLoci and it's children class
        :return: Boolean
        """
        return self.chromosome == other.chromosome and self.start <= other.end and self.end >= other.start

    @classmethod
    def create_loci(cls, string):
        u"""
        Create loci from String
        :param string: chr1:1-100:+
        :return:
        """
        temp = string.split(":")

        if len(temp) == 3:
            chromosome, sites, strand = temp
        elif len(temp) == 2:
            chromosome, sites = temp
            strand = "*"
        else:
            raise ValueError("Failed to decode genomic region: %s" % string)

        start, end = sites.split("-")

        return cls(chromosome, start, end, strand)

    def relative(self, site: int) -> int:
        return site - self.start


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
u"""
Created by ygidtu@gmail.com at 2019.12.06

Changelog:
    1. move several attributes and functions to corresponding file objects, turn this into pure data class
    2. add transform to log2, log10 or zscore transform the data while plotting
"""
from typing import Dict, Optional

import numpy as np
from scipy.stats import zscore

from sashimi.base.Junction import Junction


class ReadDepth(object):
    u"""
    Migrated from SplicePlot ReadDepth class

    add a parent class to handle all the position comparison
    """

    def __init__(self,
                 wiggle: np.array,
                 junctions_dict: Optional[Dict[Junction, int]] = None,
                 site_plus: Optional[np.array] = None,
                 site_minus: Optional[np.array] = None,
                 minus: Optional[np.array] = None,
                 junction_dict_plus: Optional[np.array] = None,
                 junction_dict_minus: Optional[np.array] = None,
                 strand_aware: bool = False):
        u"""
        init this class
        :param wiggle: a numpy.ndarray object represented the whole read coverage,
                       should be summation of plus and minus or plus
        :param junctions_dict: a dict represented the coordinate of each intron as well as frequency
        :param site_plus: a numpy.ndarray object represented the forward site coverage
        :param site_minus: a numpy.ndarray object represented the reverse site coverage
        :param minus: a numpy.ndarray object represented the reverse strand read coverage
        :param strand_aware: strand specific depth
        :param junction_dict_plus: these splice junction from plus strand
        :param junction_dict_minus: these splice junction from minus strand
        """
        self.plus = wiggle
        self.junctions_dict = junctions_dict
        self.strand_aware = strand_aware
        self.minus = abs(minus) if minus is not None else minus
        self.junction_dict_plus = junction_dict_plus
        self.junction_dict_minus = junction_dict_minus
        self.site_plus = site_plus
        self.site_minus = site_minus * -1 if site_minus is not None else site_minus

    @property
    def wiggle(self) -> np.array:
        if (self.plus is None or not self.plus.any()) and self.minus is not None:
            return self.minus

        if self.plus is not None and self.minus is not None:
            return self.plus + self.minus

        return self.plus

    @property
    def max(self) -> float:
        return max(self.wiggle, default=0)

    def __add__(self, other):

        """
            __add__ allows two ReadDepth objects to be added together using the + symbol

            Both self and other must have the same low and high attributes

            return value:
                A new ReadDepth object containing the sum of the two original ReadDepth objects
        """

        if self.wiggle is not None and other.wiggle is not None:
            if len(self.wiggle) == len(other.wiggle):
                junctions = self.junctions_dict if self.junctions_dict else {}
                if other.junctions_dict:
                    for i, j in other.junctions_dict.items():
                        if i in junctions.keys():
                            junctions[i] += j
                        else:
                            junctions[i] = j

                minus = None
                if self.minus is not None and other.minus is not None:
                    minus = self.minus + other.minus
                elif self.minus is None and other.minus is not None:
                    minus = other.minus
                elif self.minus is not None and other.minus is None:
                    minus = self.minus

                return ReadDepth(
                    self.plus + other.plus,
                    junctions_dict=junctions,
                    minus=minus
                )
        elif self.wiggle is None:
            return other
        else:
            return self

    def curr_height(self, pos: int) -> float:
        if self.minus is None:
            return self.plus[pos]
        return self.plus[pos] + self.minus[pos]

    def curr_max(self, pos: int) -> float:
        return self.plus[pos]

    def curr_min(self, pos: int) -> float:
        return self.minus[pos] if self.minus is not None else 0

    def add_customized_junctions(self, other):
        u"""
        Add customized junctions to plot
        :param other:
        :return:
        """
        new_junctions_dict = {}

        for key, value in self.junctions_dict.items():
            if key in other.junctions_dict:
                new_junctions_dict[key] = value + other.junctions_dict[key]
            else:
                new_junctions_dict[key] = value

        for key, value in list(other.junctions_dict.items()):
            if key not in self.junctions_dict:
                new_junctions_dict[key] = value

        self.junctions_dict = new_junctions_dict
        return new_junctions_dict

    def transform(self, log_trans: str):
        funcs = {"10": np.log10, "2": np.log2, "zscore": zscore, "e": np.log}

        if log_trans in funcs.keys():
            if self.plus is not None:
                self.plus = funcs[log_trans](self.plus + 1)

            if self.minus is not None:
                self.minus = funcs[log_trans](self.minus + 1)

    def normalize(self, size_factor: float):
        self.plus = np.divide(self.plus, size_factor) # * 100
        if self.minus is not None:
            self.minus = np.divide(self.minus, size_factor)


if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
u"""
Created at 2022.05.31

This script contains all the command line parameters
"""
import gzip
import os
import sys
from multiprocessing import cpu_count
from typing import Optional, Dict, Set, Tuple

import click
import matplotlib as mpl
from click_option_group import optgroup
from loguru import logger

from sashimi.base.GenomicLoci import GenomicLoci
from sashimi.conf.config import CLUSTERING_METHOD, COLORS, COLORMAP, DISTANCE_METRIC, IMAGE_TYPE
from sashimi.file.ATAC import ATAC
from sashimi.plot import Plot

__version__ = "0.1.2"
__author__ = "ygidtu & Ran Zhou"
__email__ = "ygidtu@gmail.com"


def decode_region(region: str):
    regions = region.split(":")

    if len(regions) < 3:
        strand = "+"
    else:
        strand = regions[-1]

    sites = [int(x) for x in regions[1].split("-")]
    return GenomicLoci(regions[0], sites[0], sites[1], strand)


class FileList(object):
    def __init__(self,
                 path: str,
                 category: str,
                 color: str = "black",
                 label: Optional[str] = None,
                 group: Optional[str] = None,
                 exon_focus: Optional[str] = None,
                 library: str = "fru",
                 trans: Optional[str] = None,
                 depth: int = 30000):

        self.path = os.path.abspath(os.path.expanduser(path))

        if not os.path.exists(self.path):
            raise FileNotFoundError(f"{self.path} not found.")

        self.label = label if label else os.path.basename(path)
        self.group = group
        self.color = color
        self.category = category
        self.exon_focus = exon_focus
        self.library = library
        self.trans = trans
        self.depth = depth

    @property
    def name(self) -> str:
        return os.path.basename(self.path)

    def __str__(self):
        return f"path: {self.path} \nlabel: {self.label} \ngroup: {self.group} \n" \
               f"color: {self.color} \ncategory: {self.category} \nlibrary: {self.library}"


def load_barcodes(barcode: str) -> Tuple[Dict[str, Dict[str, Set[str]]], Dict[str, str]]:
    u"""
    as name says
    :param barcode: the path to barcode file
    """
    r = gzip.open(barcode, "rt") if barcode.endswith(".gz") else open(barcode, "r")
    res = {}
    colors = {}
    for line in r:
        line = line.strip().split()
        if len(line) >= 4:
            key, bc, group, color = line[:4]
            colors[group] = color
        elif len(line) == 3:
            key, bc, group = line[:3]
        elif len(line) == 2:
            key, bc = line
            group = ""
        else:
            continue

        if key not in res.keys():
            res[key] = {}

        if group not in res[key]:
            res[key][group] = set()

        res[key][group].add(bc)

    r.close()
    return res, colors


def __read_iter__(path):
    with open(path) as r:
        for idx, line in enumerate(r):
            if line.startswith("#"):
                continue
            line = line.split()
            if not line:
                continue
            yield idx, line


def process_file_list(infile: str, category: str = "density"):
    u"""
    Process and check the file list format_
    :param infile: path to input file list
    :param category: the image type of file list used for
    """

    try:
        if category in ["density"]:
            for idx, line in __read_iter__(infile):
                path, category = line[0], line[1]

                if category not in ["bam", "bigwig", "bw", "depth", "igv", "atac"]:
                    raise ValueError(f"{category} is not supported in density plot.")

                if len(line) < 3:
                    yield FileList(path=path, category=category, color=COLORS[idx % len(COLORS)])
                elif len(line) < 4:
                    yield FileList(path=path, category=category, color=COLORS[idx % len(COLORS)], label=line[2])
                elif len(line) < 5:
                    yield FileList(path=path, category=category, color=line[3], label=line[2])
                elif len(line) < 6:
                    yield FileList(path=path, category=category, color=line[3], label=line[2], library=line[4])
                else:
                    yield FileList(path=path, category=category, color=line[3], label=line[2], library=line[4],
                                   depth=line[5])
        elif category in ["heatmap"]:
            groups = {}
            for idx, line in __read_iter__(infile):
                path, category = line[0], line[1]

                if category not in ["bam", "bigwig", "bw", "depth", "atac"]:
                    raise ValueError(f"{category} is not supported in heatmap plot.")

                if len(line) < 3:
                    yield FileList(path=path, category=category, color=COLORMAP[0])
                elif len(line) < 4:
                    groups[line[2]] = 0
                    yield FileList(path=path, category=category,
                                   color=COLORMAP[len(groups) % len(COLORMAP)], group=line[2])
                else:
                    groups[line[2]] = 0
                    yield FileList(path=path, category=category,
                                   color=line[3], group=line[2], library=line[4] if len(line) > 4 else "fru")
        elif category in ["line"]:
            groups = {}
            for idx, line in __read_iter__(infile):
                path, category = line[0], line[1]

                if category not in ["bam", "bigwig", "bw", "depth"]:
                    raise ValueError(f"{category} is not supported in density plot.")

                if len(line) < 3:
                    yield FileList(path=path, category=category, color=COLORS[idx % len(COLORS)])
                elif len(line) < 4:
                    if line[2] not in groups:
                        groups[line[2]] = 0
                    groups[line[2]] += 1
                    yield FileList(path=path, category=category,
                                   color=COLORS[groups[line[2]] % len(COLORS)], group=line[2])
                elif len(line) < 5:
                    if line[2] not in groups:
                        groups[line[2]] = 0
                    groups[line[2]] += 1
                    yield FileList(path=path, category=category, label=line[3],
                                   color=COLORS[groups[line[2]] % len(COLORS)], group=line[2])
                elif len(line) < 6:
                    if line[2] not in groups:
                        groups[line[2]] = 0
                    groups[line[2]] += 1
                    yield FileList(path=path, category=category, label=line[3],
                                   color=line[4], group=line[2])
        elif category in ["interval"]:
            for idx, line in __read_iter__(infile):
                if len(line) < 2:
                    yield FileList(path=line[0], category="interval")
                else:
                    yield FileList(path=line[0], category="interval", label=line[1])
        elif category in ["igv"]:
            for idx, line in __read_iter__(infile):
                path, category = line[0], line[1]

                if category not in ["bam", "bigwig", "bw", "depth", "igv"]:
                    raise ValueError(f"{category} is not supported in density plot.")

                if len(line) < 3:
                    yield FileList(path=path, category=category, color=COLORS[idx % len(COLORS)])
                elif len(line) < 4:
                    yield FileList(path=path, category=category, color=COLORS[idx % len(COLORS)], label=line[2])
                elif len(line) < 5:
                    yield FileList(path=path, category=category, color=line[3], label=line[2])
                else:
                    yield FileList(path=path, category=category, color=line[3], label=line[2], exon_focus=line[4])
        elif category in ["hic"]:
            for idx, line in __read_iter__(infile):
                path, category = line[0], line[1]
                if len(line) < 3:
                    yield FileList(path=path, category=category)
                elif len(line) < 4:
                    yield FileList(path=path, category=category, label=line[2])
                elif len(line) < 5:
                    yield FileList(path=path, category=category, )
                else:
                    yield FileList(path=path, category=category, label=line[2], color=line[3], trans=line[4])
    except FileNotFoundError as err:
        logger.error(f"{infile} -> {err}")
        exit(1)

    return None


@click.command(context_settings=dict(help_option_names=['-h', '--help']), )
@click.version_option(__version__, message="Current version %(version)s")
@click.option("--debug", is_flag=True, help="enable debug level log")
@click.option("-e", "--event", type=click.STRING, required=True,
              help="Event range eg: chr1:100-200:+")
@optgroup.group("Common input files configuration")
@optgroup.option("--color-factor", default=1, type=click.IntRange(min=1),
                 help="Index of column with color levels (1-based); "
                      "NOTE: LUAD|red -> LUAD while be labeled in plots and red while be the fill color",
                 show_default=True)
@optgroup.option("--barcode", type=click.Path(exists=True), show_default=True,
                 help="Path to barcode list file, At list two columns were required, "
                      "- 1st The name of bam file, not the alias of bam; \b"
                      "- 2nd the barcode; \b"
                      "- 3rd The group label, optional; \b"
                      "- 4th The color of each cell type, default using the color of corresponding bam file.\n")
@optgroup.option("--barcode-tag", type=click.STRING, default="CB", show_default=True,
                 help="The default cell barcode tag label")
@optgroup.option("--umi-tag", type=click.STRING, default="UB", show_default=True,
                 help="The default UMI barcode tag label")
@optgroup.option("-p", "--process", type=click.IntRange(min=1, max=cpu_count()), default=1,
                 help="How many cpu to use")
@optgroup.option("--group-by-cell", type=click.BOOL, is_flag=True, help="Group by cell types in density/line plot")
@optgroup.option("--remove-duplicate-umi", type=click.BOOL, is_flag=True, help="Drop duplicated UMIs by barcode")
@optgroup.group("Output settings")
@optgroup.option("-o", "--output", type=click.Path(),
                 help="Path to output graph file", show_default=True)
@optgroup.option("-d", "--dpi", default=300, type=click.IntRange(min=1, clamp=True),
                 help="The resolution of output file", show_default=True)
@optgroup.option("--raster", is_flag=True, show_default=True,
                 help="The would convert heatmap and site plot to raster image "
                      "(speed up rendering and produce smaller files), only affects pdf, svg and PS")
@optgroup.option("--height", default=1, type=float,
                 help="The height of output file, default adjust image height by content", show_default=True)
@optgroup.option("--width", default=10, type=click.IntRange(min=0, clamp=True),
                 help="The width of output file, default adjust image width by content", show_default=True)
@optgroup.option("--backend", type=click.STRING, default="Agg", help="Recommended backend", show_default=True)
@optgroup.group("Reference settings")
@optgroup.option("-r", "--reference", type=click.Path(exists=True),
                 help="Path to gtf file, both transcript and exon tags are necessary")
@optgroup.option("--interval", type=click.Path(exists=True),
                 help="Path to list of interval files in bed format, "
                      "1st column is path to file, 2nd column is the label [optional]")
@optgroup.option("--show-id", is_flag=True, show_default=True, help="Whether show gene id or gene name")
@optgroup.option("--show-exon-id", is_flag=True, show_default=True, help="Whether show gene id or gene name")
@optgroup.option("--no-gene", is_flag=True, type=click.BOOL, show_default=True,
                 help="Do not show gene id next to transcript id")
@optgroup.option("--domain", default=False, is_flag=True, type=click.BOOL, show_default=True,
                 help="Add domain information into reference track")
@optgroup.option("--proxy", default=None, type=click.STRING,
                 help="The http or https proxy for EBI/Uniprot requests,"
                      "if `--domain` is True, eg: http://127.0.0.1:1080")
@optgroup.option("--timeout", default=10, type=click.IntRange(min=1, clamp=True),
                 show_default=True,
                 help="The requests timeout when `--domain` is True.")
@optgroup.option("--local-domain", default="", is_flag=False, type=click.STRING, show_default=True,
                 help="Load local domain folder and load into reference track, download from "
                      "https://hgdownload.soe.ucsc.edu/gbdb/hg38/uniprot/")
@optgroup.option("--remove-empty", is_flag=True, type=click.BOOL, show_default=True,
                 help="Whether to plot empty transcript")
@optgroup.option("--transcripts-to-show", default="", show_default=True,
                 help="Which transcript to show, transcript name or id in gtf file, eg: transcript1,transcript2")
@optgroup.option("--choose-primary", is_flag=True, type=click.BOOL, show_default=True,
                 help="Whether choose primary transcript to plot.")
@optgroup.option("--ref-color", default="black", type=click.STRING,
                 show_default=True, help="The color of exons")
@optgroup.option("--intron-scale", type=click.FLOAT, default=0.5, help="The scale of intron", show_default=True)
@optgroup.option("--exon-scale", type=click.FLOAT, default=1, help="The scale of exon", show_default=True)
@optgroup.group("Density plot settings")
@optgroup.option("--density", type=click.Path(exists=True),
                 help="""
                 The path to list of input files, a tab separated text file, \b
                 - 1st column is path to input file, \b
                 - 2nd column is the file category, \b
                 - 3rd column is input file alias (optional), \b
                 - 4th column is color of input files (optional),
                 - 5th column is the library of input file (optional, only required by bam file). \n
                 """)
@optgroup.option("--customized-junction", type=click.STRING, default=None, show_default=True,
                 help="Path to junction table column name needs to be bam name or bam alias.")
@optgroup.option("--only-customized-junction", is_flag=True, show_default=True, help="Only used customized junctions.")
@optgroup.option("-t", "--threshold", default=0, type=click.IntRange(min=0, clamp=True),
                 show_default=True, help="Threshold to filter low abundance junctions")
@optgroup.option("--density-by-strand", is_flag=True, type=click.BOOL,
                 show_default=True, help="Whether to draw density plot by strand")
@optgroup.option("--show-site", is_flag=True, type=click.BOOL,
                 show_default=True, help="Whether to draw additional site plot")
@optgroup.option("--site-strand", type=click.Choice(["all", "+", "-"]), default="all", show_default=True,
                 help="Which strand kept for site plot, default use all")
@optgroup.option("--included-junctions", type=click.STRING, default=None,
                 help="The junction id for including, chr1:1-100", show_default=True)
@optgroup.option("--show-junction-num", type=click.BOOL, is_flag=True, show_default=True,
                 help="Whether to show the number of junctions")
@optgroup.option("--sc-density-height-ratio", type=float, default=1, show_default=True,
                 help="The relative height of single cell density plots")
@optgroup.group("Line plot settings")
@optgroup.option("--line", type=click.Path(exists=True),
                 help="""
                 The path to list of input files, a tab separated text file, \b
                 - 1st column is path to input file, \b
                 - 2nd column is the file category, \b
                 - 3rd column is input file group (optional), \b
                 - 4th column is input file alias (optional),
                 - 5th column is color platte of corresponding group (optional).
                 """)
@optgroup.option("--hide-legend", default=False, is_flag=True, type=click.BOOL, help="Whether to hide legend")
@optgroup.option("--legend-position", default="upper right", type=click.STRING, help="The legend position")
@optgroup.option("--legend-ncol", default=0, type=click.IntRange(min=0, clamp=True),
                 help="The number of columns of legend")
@optgroup.group("Heatmap plot settings")
@optgroup.option("--heatmap", type=click.Path(exists=True),
                 help="""
                 The path to list of input files, a tab separated text file, \b
                 - 1st column is path to input file, \b
                 - 2nd column is the file category, \b
                 - 3rd column is input file group (optional), \b
                 - 4th column is color platte of corresponding group.
                 """)
@optgroup.option("--clustering", is_flag=True, show_default=True, help="Enable clustering of the heatmap")
@optgroup.option("--clustering-method", type=click.Choice(CLUSTERING_METHOD), default="ward",
                 show_default=True, help="The clustering method for heatmap")
@optgroup.option("--distance-metric", type=click.Choice(DISTANCE_METRIC), default="euclidean",
                 show_default=True, help="The distance metric for heatmap")
@optgroup.option("--heatmap-scale", is_flag=True, show_default=True, help="Do scale on heatmap matrix.")
@optgroup.option("--heatmap-vmin", type=click.INT, show_default=True,
                 help="Minimum value to anchor the colormap, otherwise they are inferred from the data.")
@optgroup.option("--heatmap-vmax", type=click.INT, show_default=True,
                 help="Maximum value to anchor the colormap, otherwise they are inferred from the data.")
@optgroup.option("--show-row-names", is_flag=True, show_default=True, help="Show row names of heatmap")
@optgroup.option("--sc-heatmap-height-ratio", type=float, default=.2, show_default=True,
                 help="The relative height of single cell heatmap plots")
@optgroup.group("IGV settings")
@optgroup.option("--igv", type=click.Path(exists=True),
                 help="""
                 The path to list of input files, a tab separated text file, \b
                 - 1st column is path to input file, \b
                 - 2nd column is the file category, \b
                 - 3rd column is input file alias (optional), \b
                 - 4th column is color of input files (optional),\b
                 - 5th column is exon_id for sorting the reads (optional).
                 """)
@optgroup.option("--m6a", default=None, type=click.STRING,
                 help="""
                 Sashimi.py will load location information from the given tags and \b
                  then highlight the RNA m6a modification cite at individual reads. \b
                  If there are multiple m6a modification site, please add tag as follow, \b
                  234423,234450
                 """)
@optgroup.option("--polya", default=None, type=click.STRING,
                 help="""
                 Sashimi.py will load length of poly(A) from the given tags and \b
                  then visualize the poly(A) part at end of each individual reads.
                 """)
@optgroup.option("--rs", default=None, type=click.STRING,
                 help="""
                 Sashimi.py will load real strand information of each reads from the given tags and \b
                  the strand information is necessary for visualizing poly(A) part.
                 """)
@optgroup.option("--del-ratio-ignore", default=1.0,
                 type=click.FloatRange(min=0.0, max=1.0, clamp=True),
                 help="""
                 Ignore the deletion gap in nanopore or pacbio reads. \b
                 if a deletion region was smaller than (alignment length) * (del_ratio_ignore), \b
                 then the deletion gap will be filled. \b
                 currently the del_ratio_ignore was 1.0.
                 """)
@optgroup.group("HiC settings")
@optgroup.option("--hic", type=click.Path(exists=True),
                 help="""
                 The path to list of input files, a tab separated text file, \b
                 - 1st column is path to input file, \b
                 - 2nd column is the file category, \b
                 - 3rd column is input file alias (optional), \b
                 - 4th column is color of input files (optional),\b
                 - 5th column is data transform for HiC matrix, eg log1p, log2, log10 (optional).
                 """)
@optgroup.group("Additional annotation")
@optgroup.option("-f", "--genome", type=click.Path(), default=None,
                 show_default=True, help="Path to genome fasta")
@optgroup.option("--sites", default=None, type=click.STRING,
                 help="Where to plot additional indicator lines, comma separated int")
@optgroup.option("--stroke", type=click.STRING, show_default=True,
                 help="The stroke regions: start1-end1:start2-end2@color-label, "
                      "draw a stroke line at bottom, default color is red")
@optgroup.option("--link", type=click.STRING, show_default=True,
                 help="The link: start1-end1:start2-end2@color, "
                      "draw a link between two site at bottom, default color is blue")
@optgroup.option("--focus", type=click.STRING, show_default=True, help="The highlight regions: 100-200:300-400")
@optgroup.group("Motif settings")
@optgroup.option("--motif", type=click.Path(exists=True),
                 help="The path to customized bedGraph file, first three columns is chrom, start and end site, "
                      "the following 4 columns is the weight of ATCG.")
@optgroup.option("--motif-region", type=click.STRING, default="",
                 help="The region of motif to plot in start-end format", show_default=True)
@optgroup.option("--motif-width", type=click.FLOAT, default=0.8,
                 help="The width of ATCG characters", show_default=True)
@optgroup.group("Layout settings")
@optgroup.option("--n-y-ticks", default=4, type=click.IntRange(min=0, clamp=True),
                 help="The number of ticks of y-axis")
@optgroup.option("--distance-ratio", type=click.FLOAT, default=0.1,
                 help="distance between transcript label and transcript line", show_default=True)
@optgroup.option("--reference-scale", type=click.FLOAT, default=.25,
                 help="The size of reference plot in final plot", show_default=True)
@optgroup.option("--stroke-scale", type=click.FLOAT, default=.25,
                 help="The size of stroke plot in final image", show_default=True)
@optgroup.group("Overall settings")
@optgroup.option("--font-size", default=8, type=click.IntRange(min=1, clamp=True),
                 help="The font size of x, y-axis and so on")
@optgroup.option("--reverse-minus", default=False, is_flag=True, type=click.BOOL,
                 help="Whether to reverse strand of bam/reference file")
@optgroup.option("--hide-y-label", default=False, is_flag=True, type=click.BOOL,
                 help="Whether hide y-axis label")
@optgroup.option("--same-y", default=False, is_flag=True, type=click.BOOL,
                 help="Whether different sashimi/line plots shared same y-axis boundaries")
@optgroup.option('--log', type=click.Choice(["0", "2", "10", "zscore"]), default="0",
                 help="y axis log transformed, 0 -> not log transform; 2 -> log2; 10 -> log10")
@optgroup.option("--title", type=click.STRING, default=None, help="Title", show_default=True)
@optgroup.option("--font", type=click.STRING, default=None, help="Fonts", show_default=True)
def main(**kwargs):
    u"""
    Welcome to use sashimi
    \f
    """
    if not kwargs["debug"]:
        logger.remove()
        logger.add(sys.stderr, level="INFO")
        logger.level("INFO")

    # print warning info about backend
    if (kwargs["domain"] or kwargs["local_domain"]) and kwargs["backend"].lower() != "cairo":
        logger.warning(f"{kwargs['backend']} backend may have problems with small domain, "
                       f"if there is any please try cairo backend instead.")

    if kwargs["raster"] and kwargs["heatmap"] and kwargs["backend"].lower() == "cairo":
        logger.warning(f"{kwargs['backend']} backend may have problems with rasterized heatmap, "
                       f"if there is any, please try another backend instead.")

    try:
        mpl.use(kwargs["backend"])
    except ImportError as err:
        if kwargs["backend"].lower() == "cairo":
            logger.warning("Cairo backend required cairocffi installed")
            logger.warning("Switch back to Agg backend")
        else:
            logger.warning(f"backend error, switch back to Agg: {err}")
        mpl.use("Agg")

    mpl.rcParams['pdf.fonttype'] = 42

    if kwargs["font"]:
        mpl.rcParams['font.family'] = kwargs["font"]

    for k, v in kwargs.items():
        logger.debug(f"{k} => {v}")

    if kwargs["included_junctions"] is not None:
        included_junctions = set([sj.strip() for sj in kwargs["included_junctions"].split(',')])
    else:
        included_junctions = {}

    p = Plot()

    region = decode_region(kwargs["event"])
    p.set_region(region=region)
    p.add_customized_junctions(kwargs["customized_junction"])

    barcodes, sc_colors = {}, {}
    if kwargs.get("barcode") and os.path.exists(kwargs.get("barcode")):
        barcodes, sc_colors = load_barcodes(kwargs.get("barcode"))

    size_factors = {}

    # add reference
    for key in kwargs.keys():
        # print(key)
        if key in IMAGE_TYPE and kwargs[key] and os.path.exists(kwargs[key]):
            logger.debug(f"add {key} {kwargs[key]}")
            if key == "reference":
                p.set_reference(kwargs["reference"],
                                show_gene=not kwargs["no_gene"],
                                color=kwargs["ref_color"],
                                remove_empty_transcripts=kwargs["remove_empty"],
                                choose_primary=kwargs["choose_primary"],
                                font_size=kwargs["font_size"],
                                show_id=kwargs["show_id"],
                                reverse_minus=kwargs["reverse_minus"],
                                show_exon_id=kwargs["show_exon_id"],
                                transcripts=kwargs["transcripts_to_show"],
                                add_domain=kwargs["domain"],
                                local_domain=kwargs["local_domain"]
                                )
            elif key == "interval":
                for f in process_file_list(kwargs[key], key):
                    p.add_interval(f.path, f.label)
            elif key == "density":
                for f in process_file_list(kwargs[key], key):
                    if barcodes and f.name in barcodes.keys() and f.category in ["bam", "atac"]:
                        for group in barcodes[f.name].keys():
                            if kwargs["group_by_cell"] and group:
                                label = group
                            elif group:
                                label = f"{f.label} - {group}"
                            else:
                                label = f.label

                            if f.label not in size_factors.keys() and f.category == "atac":
                                logger.info(f"Indexing {f.path}")
                                size_factors[f.label] = ATAC.index(f.path, barcodes[f.name])

                            p.add_density(f.path,
                                          category=f.category,
                                          label=label,
                                          barcode=group,
                                          barcode_groups=barcodes[f.name],
                                          barcode_tag=kwargs["barcode_tag"],
                                          umi_tag=kwargs["umi_tag"],
                                          library=f.library,
                                          size_factor=size_factors.get(f.label),
                                          color=sc_colors.get(group, f.color),
                                          font_size=kwargs["font_size"],
                                          show_junction_number=kwargs["show_junction_num"],
                                          n_y_ticks=kwargs["n_y_ticks"],
                                          show_y_label=not kwargs["hide_y_label"],
                                          show_site_plot=kwargs["show_site"],
                                          strand_choice=kwargs["site_strand"],
                                          density_by_strand=kwargs["density_by_strand"],
                                          only_customized_junction=kwargs["only_customized_junction"],
                                          log_trans=kwargs["log"])
                    elif f.category != "atac":
                        p.add_density(f.path,
                                      category=f.category,
                                      label=f.label,
                                      barcode_tag=kwargs["barcode_tag"],
                                      umi_tag=kwargs["umi_tag"],
                                      library=f.library,
                                      color=f.color,
                                      font_size=kwargs["font_size"],
                                      show_junction_number=kwargs["show_junction_num"],
                                      n_y_ticks=kwargs["n_y_ticks"],
                                      show_y_label=not kwargs["hide_y_label"],
                                      show_site_plot=kwargs["show_site"],
                                      strand_choice=kwargs["site_strand"],
                                      density_by_strand=kwargs["density_by_strand"],
                                      log_trans=kwargs["log"])
            elif key == "heatmap":
                for f in process_file_list(kwargs[key], key):
                    if barcodes and f.name in barcodes.keys() and f.category in ["bam", "atac"]:
                        if f.label not in size_factors.keys() and f.category == "atac":
                            logger.info(f"Indexing {f.path}")
                            size_factors[f.label] = ATAC.index(f.path, barcodes[f.name])

                        for group in barcodes[f.name].keys():
                            p.add_heatmap(f.path,
                                          category=f.category,
                                          label=f"{f.label} - {group}" if group else f.label,
                                          barcode=group,
                                          barcode_groups=barcodes[f.name],
                                          group=f"{f.group} - {group}" if f.group else f.group,
                                          barcode_tag=kwargs["barcode_tag"],
                                          size_factor=size_factors.get(f.label),
                                          umi_tag=kwargs["umi_tag"],
                                          library=f.library,
                                          color=f.color,
                                          show_y_label=not kwargs["hide_y_label"],
                                          clustering=kwargs["clustering"],
                                          clustering_method=kwargs["clustering_method"],
                                          distance_metric=kwargs["distance_metric"],
                                          font_size=kwargs["font_size"],
                                          do_scale=kwargs["heatmap_scale"],
                                          vmin=kwargs["heatmap_vmin"],
                                          vmax=kwargs["heatmap_vmax"],
                                          log_trans=kwargs["log"])
                    elif f.category != "atac":
                        p.add_heatmap(f.path,
                                      category=f.category,
                                      group=f.group,
                                      label=f.label,
                                      barcode_tag=kwargs["barcode_tag"],
                                      umi_tag=kwargs["umi_tag"],
                                      library=f.library,
                                      color=f.color,
                                      show_y_label=not kwargs["hide_y_label"],
                                      clustering=kwargs["clustering"],
                                      clustering_method=kwargs["clustering_method"],
                                      distance_metric=kwargs["distance_metric"],
                                      font_size=kwargs["font_size"],
                                      show_row_names=kwargs["show_row_names"],
                                      do_scale=kwargs["heatmap_scale"],
                                      vmin=kwargs["heatmap_vmin"],
                                      vmax=kwargs["heatmap_vmax"],
                                      log_trans=kwargs["log"])
            elif key == "line":
                for f in process_file_list(kwargs[key], key):
                    if barcodes and f.name in barcodes.keys() and f.category == "bam":
                        for group in barcodes[f.name].keys():
                            if kwargs["group_by_cell"] and group:
                                label = group
                            elif group:
                                label = f"{f.label} - {group}"
                            else:
                                label = f.label
                            p.add_line(f.path,
                                       category=f.category,
                                       label=label,
                                       barcode=group,
                                       barcode_groups=barcodes,
                                       group=f.group,
                                       barcode_tag=kwargs["barcode_tag"],
                                       umi_tag=kwargs["umi_tag"],
                                       library=f.library,
                                       color=sc_colors.get(group, f.color),
                                       show_y_label=not kwargs["hide_y_label"],
                                       font_size=kwargs["font_size"],
                                       n_y_ticks=kwargs["n_y_ticks"],
                                       show_legend=not kwargs["hide_legend"],
                                       legend_position=kwargs["legend_position"],
                                       legend_ncol=kwargs["legend_ncol"],
                                       log_trans=kwargs["log"])
                    else:
                        p.add_line(f.path,
                                   category=f.category,
                                   group=f.group,
                                   label=f.label,
                                   barcode_tag=kwargs["barcode_tag"],
                                   umi_tag=kwargs["umi_tag"],
                                   library=f.library,
                                   color=f.color,
                                   show_y_label=not kwargs["hide_y_label"],
                                   font_size=kwargs["font_size"],
                                   n_y_ticks=kwargs["n_y_ticks"],
                                   show_legend=not kwargs["hide_legend"],
                                   legend_position=kwargs["legend_position"],
                                   legend_ncol=kwargs["legend_ncol"],
                                   log_trans=kwargs["log"])
            elif key == "igv":
                for f in process_file_list(kwargs[key], "igv"):
                    igv_features = {}
                    if kwargs["m6a"]:
                        igv_features.update({"m6a": kwargs["m6a"]})

                    if kwargs["polya"] and kwargs["rs"]:
                        igv_features.update({"real_strand": kwargs["rs"], "polya": kwargs["polya"]})

                    if len(igv_features) == 0:
                        igv_features = None

                    p.add_igv(f.path,
                              category=f.category,
                              label=f.label,
                              exon_color=f.color,
                              intron_color=f.color,
                              features=igv_features,
                              font_size=kwargs["font_size"],
                              n_y_ticks=kwargs["n_y_ticks"],
                              show_y_label=not kwargs["hide_y_label"],
                              deletion_ignore=True if kwargs["del_ratio_ignore"] == 1.0 else False,
                              del_ratio_ignore=kwargs["del_ratio_ignore"],
                              exon_focus=f.exon_focus
                              )
            elif key == "hic":
                for f in process_file_list(kwargs[key], "hic"):
                    p.add_hic(
                        f.path,
                        category=f.category,
                        label=f.label,
                        trans=f.trans,
                        depth=f.depth,
                        color=f.color,
                        show_legend=not kwargs["hide_legend"],
                        show_y_label=not kwargs["hide_y_label"],
                        font_size=kwargs["font_size"],
                        n_y_ticks=kwargs["n_y_ticks"]
                    )
            elif key == "motif":
                motif_region = None
                if kwargs["motif_region"]:
                    start, end = [int(x) for x in kwargs["motif_region"].split("-")]
                    motif_region = GenomicLoci(
                        region.chromosome,
                        max(start, region.start),
                        min(region.end, end),
                        region.strand)

                p.add_motif(kwargs[key], motif_region=motif_region, width=kwargs["motif_width"])
        elif key == "focus":
            p.add_focus(kwargs[key])
        elif key == "stroke":
            p.add_stroke(kwargs[key])
        elif key == "sites":
            p.add_sites(kwargs[key])
        elif key == "link":
            p.add_links(kwargs[key])

    if kwargs["group_by_cell"]:
        p.merge_by_cell()

    p.plot(
        kwargs["output"],
        width=kwargs["width"],
        height=kwargs["height"],
        dpi=kwargs["dpi"],
        raster=kwargs["raster"],
        intron_scale=kwargs["intron_scale"],
        exon_scale=kwargs["exon_scale"],
        reference_scale=kwargs["reference_scale"],
        strock_scale=kwargs["stroke_scale"],
        same_y=kwargs["same_y"],
        remove_duplicate_umi=kwargs["remove_duplicate_umi"],
        threshold=kwargs["threshold"],
        sc_height_ratio={
            "heatmap": kwargs["sc_heatmap_height_ratio"],
            "density": kwargs["sc_density_height_ratio"]
        },
        distance_between_label_axis=kwargs["distance_ratio"],
        included_junctions=included_junctions,
        n_jobs=kwargs.get("process", 1)
    )
    logger.info("DONE")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
u"""
Generate object for processing HiC matrix information

pre-process code was re-wrote based
https://github.com/deeptools/pyGenomeTracks/blob/c42e74e725d22269c33718d9f5df11e0c45c7378/pygenometracks/tracks/HiCMatrixTrack.py#L13
"""

import itertools
from typing import Optional

import numpy as np
from loguru import logger
from scipy import sparse

from sashimi.base.GenomicLoci import GenomicLoci
from sashimi.base.Readder import Reader


class HiCTrack:
    def __init__(self,
                 path: str,
                 label: str = "",
                 depth: int = 30000,
                 trans: Optional[str] = None,
                 matrix: Optional[np.ndarray] = None,
                 x_coord: Optional[np.ndarray] = None,
                 y_coord: Optional[np.ndarray] = None,
                 region: Optional[GenomicLoci] = None,
                 is_single_cell: bool = False
                 ):
        self.path = path
        self.matrix = matrix
        self.x = x_coord
        self.y = y_coord
        self.depth = depth
        self.trans = trans
        self.label = label
        self.region = region
        self.is_single_cell = is_single_cell

    @classmethod
    def create(cls,
               path: str,
               label: str,
               depth: int,
               trans: Optional[str] = False
               ):
        """
        Create a HiCTrack object for fetching interaction matrix
        :param path: the HiC file which could be one of [h5, cool / mcool / scool, hicpro, homer]
        :param label: the label of the given HiC data
        :param depth: the depth of the given HiC data, a bigger depth means big y-axis
        :param trans: log1p, log2 or log10 transform
        :return:
        """
        return cls(
            path=path,
            label=label,
            depth=depth,
            trans=trans
        )

    def load(self,
             region: GenomicLoci,
             **kwargs
             ):
        """
        Load data from the given region
        :param region: the GenomicLoci object of given region
        :return:
        """
        hic = Reader.read_hic(path=self.path, region=region)
        chromosome_id = region.chromosome
        region_start = region.start
        region_end = region.end
        try:
            chr_start_id, chr_end_id = hic.getChrBinRange(chromosome_id)
        except:
            logger.info("may be caused by the mismatch of chromosome")
            if chromosome_id.startswith("chr"):
                chromosome_id = region.chromosome.replace("chr", "") if region.chromosome != "chrM" else "MT"
            else:
                chromosome_id = "chr" + region.chromosome if region.chromosome != "MT" else "chrM"

            chr_start_id, chr_end_id = hic.getChrBinRange(chromosome_id)

        chr_start = hic.cut_intervals[chr_start_id][1]
        chr_end = hic.cut_intervals[chr_end_id - 1][2]

        start_bp = max(chr_start, region_start - self.depth)
        end_bp = min(chr_end, region_end + self.depth)

        idx = [idx for idx, x in enumerate(hic.cut_intervals)
               if x[0] == chromosome_id and x[1] >= start_bp and x[2] <= end_bp]

        start_pos = [x[1] for i, x in enumerate(hic.cut_intervals) if i in idx]
        start_pos = tuple(list(start_pos) + [hic.cut_intervals[idx[-1]][2]])

        matrix = hic.matrix[idx, :][:, idx]
        region_len = region_end - region_start

        current_depth = min(self.depth, int(region_len * 1.25))
        depth_in_bins = max(1, int(1.5 * region_len / hic.getBinSize()))

        if current_depth < self.depth:
            logger.warning(f"The depth was set to {self.depth} which is more than 125% "
                           f"of the region plotted. The depth will be set "
                           f"to {current_depth}.\n")
            # remove from matrix all data points that are not visible.
            matrix = matrix - sparse.triu(matrix, k=depth_in_bins, format='csr')

        matrix = np.asarray(matrix.todense().astype(float))

        n = matrix.shape[0]
        t = np.array([[1, 0.5], [-1, 0.5]])
        matrix_tmp = np.dot(np.array([(i[1], i[0])
                                      for i in itertools.product(start_pos[::-1],
                                                                 start_pos)]), t)

        self.x = matrix_tmp[:, 1].reshape(n + 1, n + 1)
        self.y = matrix_tmp[:, 0].reshape(n + 1, n + 1)
        self.matrix = HiCTrack.mat_trans(matrix, method=self.trans)
        self.region = region

    @staticmethod
    def mat_trans(matrix: np.ndarray, method: Optional[str] = None):
        if not method:
            return matrix
        assert method in {"log1p", "log2", "log10"}, f"couldn't recognize current transform method {method}"
        if method == "log1p":
            return np.log1p(matrix)
        elif method == "log2":
            return np.log2(matrix + 1)
        elif method == "log10":
            return np.log10(matrix + 1)


if __name__ == '__main__':
    pass



#!/usr/bin/pyton3
# -*- coding:utf-8 -*-
u"""

"""
import os

from sashimi.base.GenomicLoci import GenomicLoci
from sashimi.base.Readder import Reader


class Motif(object):
    def __init__(self,
                 path: str,
                 region: GenomicLoci,
                 ):
        u"""
        the plot region
        """
        self.path = path
        self.region = region
        self.label = ""
        self.data = {}  # List[List[base, score]]

    @classmethod
    def create(cls, path: str, region: GenomicLoci):
        assert os.path.exists(path), f"{path} is not exists."
        return cls(path=path, region=region)

    def load(self, region: GenomicLoci, **kwargs):
        data = {}
        keys = ["A", "T", "C", "G"]
        for record in Reader.read_depth(self.path, region):
            start = record[1]
            data[int(start)] = {x: float(y) for x, y in zip(keys, record[3:7])}
        self.data = data


if __name__ == '__main__':
    pass
